"""
Test Trainer class. To run this script, use the following command:

>>> torchrun --nnodes=1 --nproc_per_node=2 --rdzv_id=100 --rdzv_backend=c10d \
        --rdzv_endpoint=localhost:29400 torch_dist_trainer.py

"""

from torch import nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torchvision import transforms, datasets

from itwinai.torch.trainer import TorchTrainer


class Net(nn.Module):

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=0)


if __name__ == '__main__':
    train_set = datasets.MNIST(
        '.tmp/', train=True, download=True,
        transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]))
    val_set = datasets.MNIST(
        '.tmp/', train=False, download=True,
        transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]))
    train_dataloder = DataLoader(train_set, batch_size=32, pin_memory=True)
    validation_dataloader = DataLoader(val_set, batch_size=32, pin_memory=True)
    trainer = TorchTrainer(
        model=Net(),
        loss=nn.NLLLoss(),
        optimizer_class='torch.optim.SGD',
        optimizer_kwargs=dict(lr=1e-3),
        epochs=2,
        strategy='ddp',
        backend='nccl',
        checkpoint_every=1
    )
    trainer.execute(args=[train_dataloder, validation_dataloader])



import pytest
from itwinai.torch.config import TrainingConfiguration
from pydantic import ValidationError


def test_values_parsing():
    """Check dynamic override and creation of new entries."""
    cfg = TrainingConfiguration(
        batch_size='11',
        param_abc='11',
        param_xyz=1.1
    )
    assert cfg.batch_size == 11
    assert cfg.param_abc == '11'
    assert cfg.param_xyz == 1.1
    assert isinstance(cfg.pin_memory, bool)

    # Check dict-like getitem
    assert cfg['batch_size'] == 11


def test_illegal_override():
    """Test that illegal type override fails."""
    with pytest.raises(ValidationError) as exc_info:
        TrainingConfiguration(batch_size='hello')
    assert "batch_size" in str(exc_info.value)



"""
Command line interface for out Python application.
You can call commands from the command line.
Example

>>> $ itwinai --help

"""

# NOTE: import libs in the command"s function, not here.
# Otherwise this will slow the whole CLI.

from typing import Optional, List
from typing_extensions import Annotated
from pathlib import Path
import typer


app = typer.Typer(pretty_exceptions_enable=False)


@app.command()
def scalability_report(
    pattern: Annotated[str, typer.Option(
        help="Python pattern matching names of CSVs in sub-folders."
    )],
    plot_title: Annotated[Optional[str], typer.Option(
        help=("Plot name.")
    )] = None,
    skip_id: Annotated[Optional[int], typer.Option(
        help=("Skip epoch ID.")
    )] = None,
    archive: Annotated[Optional[str], typer.Option(
        help=("Archive name to backup the data, without extension.")
    )] = None,
):
    """
    Generate scalability report merging all CSVs containing epoch time
    records in sub-folders.

    Example:

    >>> itwinai scalability-report --pattern="^epoch.+\\.csv$" --skip-id 0 \\
    >>>     --plot-title "Some title" --archive archive_name

    """
    # TODO: add max depth and path different from CWD
    import os
    import re
    import glob
    import shutil
    import pandas as pd
    import matplotlib
    import matplotlib.pyplot as plt
    import numpy as np

    regex = re.compile(r'{}'.format(pattern))
    combined_df = pd.DataFrame()
    csv_files = []
    for root, _, files in os.walk(os.getcwd()):
        for file in files:
            if regex.match(file):
                fpath = os.path.join(root, file)
                csv_files.append(fpath)
                df = pd.read_csv(fpath)
                if skip_id is not None:
                    df = df.drop(df[df.epoch_id == skip_id].index)
                combined_df = pd.concat([combined_df, df])
    print("Merged CSV:")
    print(combined_df)

    avg_times = (
        combined_df
        .drop(columns='epoch_id')
        .groupby(['name', 'nodes'])
        .mean()
        .reset_index()
    )
    print("\nAvg over name and nodes:")
    print(avg_times.rename(columns=dict(time='avg(time)')))

    # fig, (sp_up_ax, eff_ax) = plt.subplots(1, 2, figsize=(12, 4))
    fig, sp_up_ax = plt.subplots(1, 1, figsize=(6, 4))
    if plot_title is not None:
        fig.suptitle(plot_title)

    sp_up_ax.set_yscale("log")
    sp_up_ax.set_xscale("log")

    markers = iter("ov^s*dXpD.+12348")

    series_names = sorted(set(avg_times.name.values))
    for name in series_names:
        df = avg_times[avg_times.name == name].drop(columns='name')

        # Debug
        # compute_time = [3791., 1884., 1011., 598.]
        # nodes = [1, 2, 4, 8]
        # d = {'nodes': nodes, 'time': compute_time}
        # df = pd.DataFrame(data=d)

        df["NGPUs"] = df["nodes"]*4
        # speedup
        df["Speedup - ideal"] = df["nodes"].astype(float)
        df["Speedup"] = df["time"].iloc[0] / df["time"]
        df["Nworkers"] = 1

        # efficiency
        df["Threadscaled Sim. Time / s"] = df["time"] * \
            df["nodes"] * df["Nworkers"]
        df["Efficiency"] = df["Threadscaled Sim. Time / s"].iloc[0] / \
            df["Threadscaled Sim. Time / s"]

        sp_up_ax.plot(
            df["NGPUs"].values, df["Speedup"].values,
            marker=next(markers), lw=1.0, label=name, alpha=0.7)

    sp_up_ax.plot(df["NGPUs"].values, df["Speedup - ideal"].values,
                  ls='dashed', lw=1.0, c='k', label="ideal")
    sp_up_ax.legend(ncol=1)

    sp_up_ax.set_xticks(df["NGPUs"].values)
    sp_up_ax.get_xaxis().set_major_formatter(
        matplotlib.ticker.ScalarFormatter())

    sp_up_ax.set_ylabel('Speedup')
    sp_up_ax.set_xlabel('NGPUs (4 per node)')
    sp_up_ax.grid()

    # Sort legend
    handles, labels = sp_up_ax.get_legend_handles_labels()
    order = np.argsort(labels)
    plt.legend([handles[idx] for idx in order], [labels[idx] for idx in order])

    plot_png = f"scaling_plot_{plot_title}.png"
    plt.tight_layout()
    plt.savefig(plot_png, bbox_inches='tight', format='png', dpi=300)
    print("Saved scaling plot to: ", plot_png)

    if archive is not None:
        if '/' in archive:
            raise ValueError("Archive name must NOT contain a path. "
                             f"Received: '{archive}'")
        if '.' in archive:
            raise ValueError("Archive name must NOT contain an extension. "
                             f"Received: '{archive}'")
        if os.path.isdir(archive):
            raise ValueError(f"Folder '{archive}' already exists. "
                             "Change archive name.")
        os.makedirs(archive)
        for csvfile in csv_files:
            shutil.copyfile(csvfile, os.path.join(archive,
                                                  os.path.basename(csvfile)))
        shutil.copyfile(plot_png, os.path.join(archive, plot_png))
        avg_times.to_csv(os.path.join(archive, "avg_times.csv"), index=False)
        print("Archived AVG epoch times CSV")

        # Copy SLURM logs: *.err *.out files
        if os.path.exists('logs_slurm'):
            print("Archived SLURM logs")
            shutil.copytree('logs_slurm', os.path.join(archive, 'logs_slurm'))
        # Copy other SLURM logs
        for ext in ['*.out', '*.err']:
            for file in glob.glob(ext):
                shutil.copyfile(file, os.path.join(archive, file))

        # Create archive
        archive_name = shutil.make_archive(
            base_name=archive,  # archive file name
            format='gztar',
            # root_dir='.',
            base_dir=archive  # folder path inside archive
        )
        shutil.rmtree(archive)
        print("Archived logs and plot at: ", archive_name)


@app.command()
def exec_pipeline(
    config: Annotated[Path, typer.Option(
        help="Path to the configuration file of the pipeline to execute."
    )],
    pipe_key: Annotated[str, typer.Option(
        help=("Key in the configuration file identifying "
              "the pipeline object to execute.")
    )] = "pipeline",
    steps: Annotated[Optional[str], typer.Option(
        help=("Run only some steps of the pipeline. Accepted values are "
              "indices, python slices (e.g., 0:3 or 2:10:100), and "
              "string names of steps.")
    )] = None,
    print_config: Annotated[bool, typer.Option(
        help=("Print config to be executed after overrides.")
    )] = False,
    overrides_list: Annotated[
        Optional[List[str]], typer.Option(
            "--override", "-o",
            help=(
                "Nested key to dynamically override elements in the "
                "configuration file with the "
                "corresponding new value, joined by '='. It is also possible "
                "to index elements in lists using their list index. "
                "Example: [...] "
                "-o pipeline.init_args.trainer.init_args.lr=0.001 "
                "-o pipeline.my_list.2.batch_size=64 "
            )
        )
    ] = None
):
    """Execute a pipeline from configuration file.
    Allows dynamic override of fields.
    """
    # Add working directory to python path so that the interpreter is able
    # to find the local python files imported from the pipeline file
    import os
    import sys
    import re
    from .utils import str_to_slice
    sys.path.append(os.path.dirname(config))
    sys.path.append(os.getcwd())

    # Parse and execute pipeline
    from itwinai.parser import ConfigParser
    overrides_list = overrides_list if overrides_list is not None else []
    overrides = {
        k: v for k, v
        in map(lambda x: (x.split('=')[0], x.split('=')[1]), overrides_list)
    }
    parser = ConfigParser(config=config, override_keys=overrides)
    if print_config:
        import json
        print()
        print("#="*15 + " Used configuration " + "#="*15)
        print(json.dumps(parser.config, indent=2))
        print("#="*50)
        print()
    pipeline = parser.parse_pipeline(pipeline_nested_key=pipe_key)
    if steps:
        if not re.match(r"\d+(:\d+)?(:\d+)?", steps):
            print(f"Looking for step name '{steps}'")
        else:
            steps = str_to_slice(steps)
        pipeline = pipeline[steps]
    pipeline.execute()

    # Cleanup PYTHONPATH
    sys.path.pop()
    sys.path.pop()


@app.command()
def mlflow_ui(
    path: str = typer.Option("ml-logs/", help="Path to logs storage."),
):
    """
    Visualize Mlflow logs.
    """
    import subprocess

    subprocess.run(f"mlflow ui --backend-store-uri {path}".split())


@app.command()
def mlflow_server(
    path: str = typer.Option("ml-logs/", help="Path to logs storage."),
    port: int = typer.Option(
        5000, help="Port on which the server is listening."),
):
    """
    Spawn Mlflow server.
    """
    import subprocess

    subprocess.run(
        f"mlflow server --backend-store-uri {path} --port {port}".split())


@app.command()
def kill_mlflow_server(
    port: int = typer.Option(
        5000, help="Port on which the server is listening."),
):
    """
    Kill Mlflow server.
    """
    import subprocess

    subprocess.run(
        f"kill -9 $(lsof -t -i:{port})",
        shell=True,
        check=True,
        stderr=subprocess.DEVNULL
    )


if __name__ == "__main__":
    app()



from typing import Tuple
import tensorflow.keras as keras
import tensorflow as tf

from itwinai.components import DataGetter, DataProcessor, monitor_exec


class MNISTDataGetter(DataGetter):
    def __init__(self):
        super().__init__()
        self.save_parameters(**self.locals2params(locals()))

    @monitor_exec
    def execute(self) -> Tuple:
        train, test = keras.datasets.mnist.load_data()
        return train, test


class MNISTDataPreproc(DataProcessor):
    def __init__(self, classes: int):
        super().__init__()
        self.save_parameters(**self.locals2params(locals()))
        self.classes = classes

    @monitor_exec
    def execute(
        self,
        *datasets,
    ) -> Tuple:
        options = tf.data.Options()
        options.experimental_distribute.auto_shard_policy = (
            tf.data.experimental.AutoShardPolicy.DATA)
        preprocessed = []
        for dataset in datasets:
            x, y = dataset
            y = keras.utils.to_categorical(y, self.classes)
            sliced = tf.data.Dataset.from_tensor_slices((x, y))
            sliced = sliced.with_options(options)
            preprocessed.append(sliced)
        return tuple(preprocessed)



"""Base TensorFlow trainer module."""
from typing import Dict, Any, Optional, List, Union, Tuple

from jsonargparse import ArgumentParser
import tensorflow as tf
from tensorflow.data import Dataset
from keras.callbacks import Callback
import keras

from ..components import Trainer, monitor_exec
from itwinai.tensorflow.distributed import get_strategy


def _import_class(name):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def _instance_from_dict(obj_dict: Dict, fail_untyped: bool = True) -> Any:
    if isinstance(obj_dict, dict) and obj_dict.get('class_path') is not None:
        # obj_dict is a dictionary with a structure compliant with
        # jsonargparse
        obj_class = _import_class(obj_dict["class_path"])
        parser = ArgumentParser()
        parser.add_subclass_arguments(
            obj_class, "object", fail_untyped=fail_untyped)
        obj_dict = {"object": obj_dict}
        return parser.instantiate_classes(obj_dict).object

    raise ValueError(
        "Unable to instantiate object with this "
        f"dict configuration: {obj_dict}.\nIt should have "
        "valid 'class_path' and 'init_args' fields"
    )


class TensorflowTrainer(Trainer):
    """Trains a Keras model.

        Args:
            epochs (int): number of training epochs.
            micro_batch_size (int): per-worker batch size. Equals macro batch
                size when not running distributed.
            shuffle_buffer (Optional[int], optional): if given, shuffles
                dataset using a buffer of given size. See
                ``tf.data.Dataset.shuffle``. Defaults to None.
            callbacks (Optional[List], optional): list fo Keras callbacks.
                Can be a list of dictionary configurations. Defaults to None.
            model_config (Optional[Dict], optional): model configuration. If
                given, a model is instantiated from this configuration.
                Defaults to None.
            model_compile_config (Optional[Dict], optional): configuration for
                ``keras.Model.compile``. Defaults to None.
            rnd_seed (Optional[int], optional): random seed. Defaults to None.
            verbose (Union[str, int], optional): verbosity level for
                ``keras.Model.fit``. Defaults to 'auto'.
    """

    #: TensorFlow distributed strategy.
    strategy: tf.distribute.Strategy
    #: Total number of workers in distributed strategy.
    num_workers: int
    #: List of Keras callbacks. Defaults to None.
    callbacks: Optional[List] = None
    #: Total number of training epochs.
    epochs: int
    #: Buffer used to shuffle dataset. Defaults to None.
    shuffle_buffer: Optional[int] = None
    #: Per-worker batch size (when distributed).
    micro_batch_size: int
    #: Total batch size. When distributed, it is the sum of
    #: ``micro_batch_size`` across all workers.
    macro_batch_size: int
    #: Random seed for reproducibility. Defaults to None.
    rnd_seed: Optional[int] = None

    def __init__(
        self,
        epochs: int,
        micro_batch_size: int,
        shuffle_buffer: Optional[int] = None,
        callbacks: Optional[List[Union[Dict, Callback]]] = None,
        model_config: Optional[Dict] = None,
        model_compile_config: Optional[Dict] = None,
        rnd_seed: Optional[int] = None,
        verbose: Union[str, int] = 'auto'
    ):
        super().__init__()
        self.save_parameters(**self.locals2params(locals()))
        self.epochs = epochs
        self.micro_batch_size = micro_batch_size
        self.shuffle_buffer = shuffle_buffer
        self.rnd_seed = rnd_seed
        self.verbose = verbose
        if callbacks is not None:
            self.callbacks = self.instantiate_callbacks(callbacks)
        else:
            self.callbacks = []

        # Distributed strategy
        self.strategy, self.num_workers = get_strategy()
        print(
            f"Distributed strategy is working with: {self.num_workers} devices"
        )
        self.macro_batch_size = self.micro_batch_size * self.num_workers

        # Compile model from configuration, if given
        if model_config is not None and model_compile_config is not None:
            with self.strategy.scope():
                self.model: tf.keras.Model = _instance_from_dict(model_config)
                model_compile_config = self.instantiate_compile_conf(
                    model_compile_config
                )
                self.model.compile(**model_compile_config)
        else:
            print(
                "Either model_config or model_compile_config were not given. "
                "Skipping automatic model compilation."
            )

    @staticmethod
    def instantiate_compile_conf(model_compile_config: Dict) -> Dict[str, Any]:
        """Instantiate fields of Keras ``model.compile()`` from
        their dictionary serialization.

        Args:
            model_compile_config (Dict): fields of Keras ``model.compile()``
                serialized as dictionary.

        Returns:
            Dict[str, Any]: dictionary mapping compile argument names to
            the instantiated objects.
        """
        final_conf = {}
        for item_name, item in model_compile_config.items():
            if isinstance(item, dict):
                item = _instance_from_dict(item)
            final_conf[item_name] = item
        return final_conf

    @staticmethod
    def instantiate_callbacks(
        callbacks: List[Union[Dict, Callback]]
    ) -> List[Callback]:
        """Instantiate Keras callbacks from dictionaries.

        Args:
            callbacks (List[Union[Dict, Callback]]): list of Keras callbacks
            in serialized as dictionary.

        Returns:
            List[Callback]: list of instantiated callbacks.
        """
        final_callbacks = []
        for item in callbacks:
            if isinstance(item, dict):
                # Not all constructor args in keras callbacks
                # are typed!
                item = _instance_from_dict(item, fail_untyped=False)
            final_callbacks.append(item)
        return final_callbacks

    @monitor_exec
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset,
        test_dataset: Optional[Dataset] = None
    ) -> Tuple[Dataset, Dataset, Dataset, keras.Model]:
        """Run training. Users should override this method.

        Args:
            train_dataset (Dataset): train dataset of type
                ``tensorflow.data.Dataset``.
            validation_dataset (Dataset): validation dataset of type
                ``tensorflow.data.Dataset``.
            test_dataset (Optional[Dataset], optional): test dataset
                of type ``tensorflow.data.Dataset``. Defaults to None.

        Returns:
            Tuple[Dataset, Dataset, Dataset, keras.Model]: tuple of
            train_dataset, validation_dataset, test_dataset, and trained
            Keras model.
        """

        print(f"len(train_dataset): {len(train_dataset)}")
        print(f"len(validation_dataset): {len(validation_dataset)}")
        print("micro_batch_size: ", self.micro_batch_size, flush=True)
        print("macro_batch_size: ", self.macro_batch_size, flush=True)

        # Shuffle dataset
        if self.shuffle_buffer:
            train_ds = train_dataset.shuffle(
                self.shuffle_buffer, seed=self.rnd_seed)
            valid_ds = validation_dataset.shuffle(
                self.shuffle_buffer, seed=self.rnd_seed)
        else:
            train_ds = train_dataset
            valid_ds = validation_dataset

        # Set batch size to the dataset and repeat
        train_ds = train_ds.batch(
            self.macro_batch_size, drop_remainder=True,
            num_parallel_calls=tf.data.AUTOTUNE
        ).repeat(self.epochs)
        valid_ds = valid_ds.batch(
            self.macro_batch_size, drop_remainder=True,
            num_parallel_calls=tf.data.AUTOTUNE
        ).repeat(self.epochs)

        print(f"len(train_ds): {len(train_ds)}")
        print(f"len(valid_ds): {len(valid_ds)}")

        # Distribute datasets among strategy's replica
        dist_train_dataset = self.strategy.experimental_distribute_dataset(
            train_ds
        )
        dist_valid_dataset = self.strategy.experimental_distribute_dataset(
            valid_ds
        )

        print(f"len(dist_train_dataset): {len(train_ds)}")
        print(f"len(dist_train_dataset): {len(valid_ds)}")

        # Compute the steps per epoch for train and valid
        steps_per_epoch = len(train_dataset) // self.macro_batch_size
        validation_steps = len(validation_dataset) // self.macro_batch_size

        print(f"steps_per_epoch: {steps_per_epoch}")
        print(f"validation_steps: {validation_steps}")

        #####################################################################
        # Instantiate here model, optimizer, loss under the strategy scope, #
        # if not done previously through `model_compile_config` and         #
        # `model_config` !                                                  #
        # Always remember that they should be instantiated under the        #
        # distributed strategy scope: ``with self.strategy.scope():``       #
        #                                                                   #
        # Example:                                                          #
        # with self.strategy.scope():                                       #
        #   model = tf.keras.Sequential(...)                                #
        #   optimizer = rf.keras.optimizers.Adam(...)                       #
        #   loss = tf.keras.losses.BinaryCrossentropy(...)                  #
        #####################################################################

        # Train the model
        self.model.fit(
            dist_train_dataset,
            validation_data=dist_valid_dataset,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            epochs=self.epochs,
            callbacks=self.callbacks,
            verbose=self.verbose
        )
        print("Training completed")
        return train_dataset, validation_dataset, test_dataset, self.model



"""
Utils to create a synthetic dataset to train generative models with.

The section is split in two parts:
- 1) **TimeSeries Dataset** Generation of a gwpy TimeSeries dataset
    making use of random noise and sinusoidal functions
- 2) **Q-plot Dataset** Conversion of TimeSeries dataset into 2D
    Image dataset making use of q_transform

The dataset used to train the NN with is created as a 2D images. Note that
you do not need to run the two sections each time, but can rather save
the dataset after creating it once and loading it at the beginning of
Process Data that to save time.
"""


import pandas as pd
import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
import multiprocessing
from gwpy.timeseries import TimeSeries
import matplotlib.transforms as mtrans


def generate_dataset_aux_channels(
        rows, columns, duration=10, sample_rate=500,
        num_waves_range=(10, 15), noise_amplitude=0.1
):
    """
    Generate a Pandas DataFrame with randomly generated smooth sine wave time
    series with added smooth random noise.

    Parameters:
        - rows (int): Number of rows in the DataFrame.
        - columns (int): Number of columns in the DataFrame.
        - duration (float): Duration of the time series data in seconds
            (default is 6 seconds).
        - sample_rate (int): Sampling rate of the time series data in samples
            per second (default is 500 samples per second).
        - num_waves_range (tuple): Range for the random number of sine waves
            to be generated for each time series.
            Format: (min_num_waves, max_num_waves) (default is (10, 15)).
        - noise_amplitude (float): Amplitude of the smooth random noise added
            to the time series data (default is 0.1).

    Returns:
        - DataFrame: Pandas DataFrame containing the generated time series
            data.
    """
    # Generate time array
    times = np.linspace(0, duration, duration * sample_rate)

    # Initialize an empty list to store individual DataFrames for each row
    dfs = []
    # columns_list = [f'Aux_{i+1}' for i in range(columns)]

    for index in range(rows):
        df_dict = {}
        for col in range(columns):
            # Initialize an array to store the generated wave data for this row
            wave_data = np.zeros(len(times))
            # Determine the number of sine waves to generate for this column
            # randomly
            num_waves = np.random.randint(*num_waves_range)

            # Generate each sine wave
            for _ in range(num_waves):
                # Randomly generate parameters for the sine wave (amplitude,
                # frequency, phase)
                amplitude = np.random.uniform(0.5, 2.0)
                frequency = np.random.uniform(0.5, 5.0)
                phase = np.random.uniform(0, 2*np.pi)

                # Generate the sine wave and add it to the wave_data
                wave_data += amplitude * \
                    np.sin(2 * np.pi * frequency * times + phase)

            # Add smooth random noise to the wave data
            smooth_noise = np.random.normal(0, noise_amplitude, len(times))
            wave_data += smooth_noise

            # Create a TimeSeries object from the wave data
            ts = TimeSeries(wave_data, t0=0, dt=1/sample_rate)
            df_dict[f'Aux_{col+1}'] = [ts]

        # Create a DataFrame with the TimeSeries
        df_row = pd.DataFrame(df_dict)

        # Append the DataFrame to the list
        dfs.append(df_row)

    # Concatenate the list of DataFrames into a single DataFrame along
    # rows axis
    df = pd.concat(dfs, ignore_index=True, axis=0)

    return df


def generate_dataset_main_channel(input_df, weights=None, noise_amplitude=0.1):
    """
    Generate a dataset where each row of a single column is a weighted linear
    combination of the entries
    in the corresponding row in the input DataFrame plus random noise.

    Parameters:
        - input_df (DataFrame): Input DataFrame generated by
            ``generate_smooth_noisy_sine_wave_series``.
        - weights (array-like): Optional weights for each entry in the row.
            If None, random weights are generated (default is None).
        - noise_amplitude (float): Amplitude of the random noise added to the
            linear combination (default is 0.1).

    Returns:
        - DataFrame: Pandas DataFrame containing the generated linear
            combination dataset.
    """
    dt = input_df.iloc[0, 0].dt
    # Initialize an empty list to store the linear combination values
    linear_combination_data = []

    # Generate random weights if not provided
    if weights is None:
        # randomly generate weights in range [0.5,1.5]
        weights = np.random.rand(len(list(input_df.columns))) + 0.5
    print(weights)
    # Iterate over rows of the input DataFrame
    for index, row in input_df.iterrows():

        # Compute the weighted linear combination of the row values
        linear_combination = np.sum(row * weights)

        # Add random noise to the linear combination
        linear_combination += np.random.normal(0, noise_amplitude)

        # Append the result to the list
        linear_combination_data.append(
            [TimeSeries(linear_combination, dt=dt, t0=0)])

    # Create a DataFrame with the linear combination data
    linear_combination_df = pd.DataFrame(
        linear_combination_data, columns=['Main'])

    return linear_combination_df


def extract_peak_frequency(hq):
    """
    Calculates peak frequency (and relative time) of a given qplot

    Input:
    -hq (gwpy.Spectrgram) : Qtransform

    Return:
    -index_time (int) : time index in qtransform relative to peak frequency
    -index_freq (int) : frequency index in qtransorm relative to peak frequency
    """

    # Calculate peak frequency, time and energy density
    index_flat = np.argmax(hq.value)

    # Convert the flattened index to 2D index
    index_time, index_freq = np.unravel_index(index_flat, hq.shape)
    # /converting_factor_frequency
    # peak_freq = hq.frequencies.value[index_freq]
    # peak_value = np.max(hq.value)
    # peak_time = hq.times.value[index_time]

    return index_time, index_freq


def cut_image(qplot_array, index_freq, index_time, square_size=64):
    """
    Cut qplot as square_size X square_size 2D np.array centered at peak
    frequency and corresponding time

    Input:
    - qplot_array (np.array) : qplot relative to the whole TimeSeries
    - index_time (int) : time index in qtransform relative to peak frequency
    - index_freq (int) : frequency index in qtransorm relative to peak
        frequency
    - square_size (int) : Size in pixels of qplot image (default 64)

    Return:
    - subarray (np.array) : qplot cutted as square_size X square_size np.array

    """

    center_x = index_time
    center_y = index_freq  # Replace with the actual y-coordinate

    original_width, original_height = qplot_array.shape

    # Calculate the starting and ending indices for the subarray
    start_x = max(center_x - square_size // 2, 0)
    # print(start_x)
    end_x = min(start_x + square_size, original_width)
    # print(end_x)
    start_y = max(center_y - square_size // 2, 0)
    # print(start_y)
    end_y = min(start_y + square_size, original_height)
    # print(end_y)
    # start_y=0
    # end_y=square_size

    # Adjust indices if needed to make sure the resulting subarray is
    # (square_size X square_size)
    if end_x - start_x < square_size:
        diff_x = square_size - (end_x - start_x)
        if end_x < original_width:
            end_x += diff_x
        else:
            start_x -= diff_x
    if end_y - start_y < square_size:
        diff_y = square_size - (end_y - start_y)
        if end_y < original_height:
            end_y += diff_y
        else:
            start_y -= diff_y

    subarray = qplot_array[start_x:end_x, start_y:end_y]
    # print(subarray.shape)
    # print(type(subarray))

    return subarray


def process_image(row, row_idx, channels, square_size):
    """
    Processes df's row to generate qplot images

    Input:
    - row (pd.Series) : row of TimeSeries Dataframe
    - row_idx (int) : index relative to row in DataFrame
    - channels (list): Name of columns of DataFrame
    - square_size (int): Size in pixels of qplot image

    Return:
    df_row (DataFrame): Row containing qplot images as 2D np.array
    """

    # res_list = []
    df_row = pd.DataFrame(columns=channels)
    for i, channel in enumerate(channels):

        qplot = row[channel].q_transform(frange=(10, 150))

        # calculate peak frequency and time indices for strain channel
        if i == 0:
            index_time, index_freq = extract_peak_frequency(qplot)

        # Convert the spectrogram to a NumPy array
        qplot_array = qplot.value
        qplot_array_cut = cut_image(
            qplot_array, index_freq, index_time, square_size)
        df_row[channel] = [qplot_array_cut]

    return df_row


def generate_cut_image_dataset(
        df, channels, num_processes=20, square_size=128
):
    """
    Generates qplot dataset taking pandas df containing main+aux channels as
    input. The output is a df containing qtransforms (frequency range 10-150Hz)
    in the form of square_sizexsquare_size 2d np.array.


     Parameters:
        - df (DataFrame): DataFrame containing Main and Aux channels'
            gwpy TimeSeries (Main channel is always first).
        - channels (list): Name of columns in the DataFrame.
        - num_processes (int): Number of cores for multiprocess (default 20)
        - square_size (int): Size in pixels of qplot image (default is
            500 samples per second).

    Returns:
        - DataFrame: Pandas DataFrame containing the q_transform np.array data.
    """

    print(channels)
    args = [(df.iloc[row], row, channels, square_size)
            for row in range(df.shape[0])]
    with multiprocessing.Pool(processes=num_processes) as pool:
        # Use map to pass multiple arguments to process_image
        results = list(
            tqdm(pool.starmap(process_image, args), total=len(args)))

    df = pd.concat(results, ignore_index=True)
    return df


def show_dataset(df, size, num_plots=10):
    """
    Plots qtransforms for first 4 columns in given df

    Input:
    - df (DataFrame) : DataFrame containing qtransforms in the form of 2d
        np.array
    - size (int) : square size in pixels of qplots
    - num_plots (int) : number of rows of df to make the plot for (default 10)

    Return
    - nothing, it just displays plots
    """

    ch_list = list(df.columns)
    fig, axes = plt.subplots(2*num_plots, 2, figsize=(18, 12*num_plots))
    for j in range(num_plots):

        qplt_r = np.flipud(df.iloc[j, 0].T)
        qplt_aux1 = np.flipud(df.iloc[j, 1].T)
        qplt_aux2 = np.flipud(df.iloc[j, 2].T)
        qplt_aux3 = np.flipud(df.iloc[j, 3].T)

        # Create a single subfigure with two plots
        # fig, axes = plt.subplots(2, 3, figsize=(18, 12))

        # Plot for Real
        im_r = axes[2*j, 0].imshow(
            qplt_r, aspect='auto',
            extent=[0, size, 0, size], vmin=0, vmax=25)
        axes[2*j, 0].set_title(ch_list[0])
        axes[2*j, 0].set_xlabel('Time')
        axes[2*j, 0].set_ylabel('Frequency')
        fig.colorbar(im_r, ax=axes[2*j, 0])  # Add colorbar for Real

        # Plot for aux
        im_g = axes[2*j, 1].imshow(
            qplt_aux1, aspect='auto',
            extent=[0, size, 0, size], vmin=0, vmax=25)
        axes[2*j, 1].set_title(ch_list[1])
        axes[2*j, 1].set_xlabel('Time')
        axes[2*j, 1].set_ylabel('Frequency')
        fig.colorbar(im_g, ax=axes[2*j, 1])  # Add colorbar for Generated
        #
        im_g = axes[2*j+1, 0].imshow(
            qplt_aux2, aspect='auto',
            extent=[0, size, 0, size], vmin=0, vmax=25)
        axes[2*j+1, 0].set_title(ch_list[2])
        axes[2*j+1, 0].set_xlabel('Time')
        axes[2*j+1, 0].set_ylabel('Frequency')
        fig.colorbar(im_g, ax=axes[2*j+1, 0])  # Add colorbar for Generated
        #
        im_g = axes[2*j+1, 1].imshow(
            qplt_aux3, aspect='auto',
            extent=[0, size, 0, size], vmin=0, vmax=25)
        axes[2*j+1, 1].set_title(ch_list[3])
        axes[2*j+1, 1].set_xlabel('Time')
        axes[2*j+1, 1].set_ylabel('Frequency')
        fig.colorbar(im_g, ax=axes[2*j+1, 1])  # Add colorbar for Generated

        # Get the bounding boxes of the axes including text decorations
        r = fig.canvas.get_renderer()

        def get_bbox(ax): return ax.get_tightbbox(
            r).transformed(fig.transFigure.inverted())
        bboxes = np.array(list(map(get_bbox, axes.flat)),
                          mtrans.Bbox).reshape(axes.shape)

        # Get the minimum and maximum extent, get the coordinate half-way
        # between those
        ymax = np.array(list(map(lambda b: b.y1, bboxes.flat))
                        ).reshape(axes.shape).max(axis=1)
        ymin = np.array(list(map(lambda b: b.y0, bboxes.flat))
                        ).reshape(axes.shape).min(axis=1)
        ys = np.c_[ymax[1:], ymin[:-1]].mean(axis=1)

        # Draw a horizontal lines at those coordinates
        for y in ys[1::2]:
            line = plt.Line2D(
                [0, 1], [y, y], transform=fig.transFigure, color="black")
            fig.add_artist(line)

    # plt.savefig('very high loss qplots.png')
    plt.show()


def normalize_(data, chan=4):
    """
    Normalizes the qplot data to the range [0,1] for NN convergence purposes

    Input:
    - data (torch.Tensor) : dataset of qtransforms
    - chan (int) : number of channels in dataset (default 4)

    Return:
    - data (torch.tensor) : normalized dataset
    """
    # Compute the maximum value for each channel across all 900 tensors
    max_vals = data.view(data.shape[0], data.shape[1], -1).max(0)[0].max(
        1)[0]
    print("Maximum values for each channel across all tensors:",
          max_vals, max_vals.shape)
    # Divide each element by the maximum value of its channel
    data /= max_vals.view(1, chan, 1, 1)
    return data



from typing import Optional, Dict, Any, Union
import os
import abc

import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset

from ..utils import dynamically_import_class, clear_key
from ..components import Predictor, monitor_exec
from .type import Batch
from ..serialization import ModelLoader


class TorchModelLoader(ModelLoader):
    """Loads a torch model from somewhere.

    Args:
        model_uri (str): Can be a path on local filesystem
            or an mlflow 'locator' in the form:
            'mlflow+MLFLOW_TRACKING_URI+RUN_ID+ARTIFACT_PATH'
    """

    def __call__(self) -> nn.Module:
        """"Loads model from model URI.

        Raises:
            ValueError: if the model URI is not recognized
                or the model is not found.

        Returns:
            nn.Module: torch neural network.
        """
        if os.path.exists(self.model_uri):
            # Model is on local filesystem.
            model = torch.load(self.model_uri)
            return model.eval()

        if self.model_uri.startswith('mlflow+'):
            # Model is on an MLFLow server
            # Form is 'mlflow+MLFLOW_TRACKING_URI+RUN_ID+ARTIFACT_PATH'
            import mlflow
            from mlflow import MlflowException
            _, tracking_uri, run_id, artifact_path = self.model_uri.split('+')
            mlflow.set_tracking_uri(tracking_uri)

            # Check that run exists
            try:
                mlflow.get_run(run_id)
            except MlflowException:
                raise ValueError(f"Run ID '{run_id}' was not found!")

            # Download model weights
            ckpt_path = mlflow.artifacts.download_artifacts(
                run_id=run_id,
                artifact_path=artifact_path,
                dst_path='tmp/',
                tracking_uri=mlflow.get_tracking_uri()
            )
            model = torch.load(ckpt_path)
            return model.eval()

        raise ValueError(
            'Unrecognized model URI: model may not be there! '
            f'Received model URI: {self.model_uri}'
        )


class TorchPredictor(Predictor):
    """Applies a pre-trained torch model to unseen data."""

    #: Pre-trained PyTorch model used to make predictions.
    model: nn.Module = None
    #: ``Dataset`` on which to make predictions (ML inference).
    test_dataset: Dataset
    #: ``DataLoader`` for test dataset.
    test_dataloader: DataLoader = None

    def __init__(
        self,
        model: Union[nn.Module, ModelLoader],
        test_dataloader_class: str = 'torch.utils.data.DataLoader',
        test_dataloader_kwargs: Optional[Dict] = None,
        name: str = None
    ) -> None:
        super().__init__(model=model, name=name)
        self.save_parameters(**self.locals2params(locals()))
        self.model = self.model.eval()

        # Train and validation dataloaders
        self.test_dataloader_class = dynamically_import_class(
            test_dataloader_class
        )
        test_dataloader_kwargs = (
            test_dataloader_kwargs
            if test_dataloader_kwargs is not None else {}
        )
        self.test_dataloader_kwargs = clear_key(
            test_dataloader_kwargs, 'train_dataloader_kwargs', 'dataset'
        )

    @monitor_exec
    def execute(
        self,
        test_dataset: Dataset,
        model: nn.Module = None,
    ) -> Dict[str, Any]:
        """Applies a torch model to a dataset for inference.

        Args:
            test_dataset (Dataset[str, Any]): each item in this dataset is a
                couple (item_unique_id, item)
            model (nn.Module, optional): torch model. Overrides the existing
                model, if given. Defaults to None.

        Returns:
            Dict[str, Any]: maps each item ID to the corresponding predicted
                value(s).
        """
        if model is not None:
            # Overrides existing "internal" model
            self.model = model

        test_dataloader = self.test_dataloader_class(
            test_dataset, **self.test_dataloader_kwargs
        )

        all_predictions = dict()
        for samples_ids, samples in test_dataloader:
            with torch.no_grad():
                pred = self.model(samples)
            pred = self.transform_predictions(pred)
            for idx, pre in zip(samples_ids, pred):
                # For each item in the batch
                if pre.numel() == 1:
                    pre = pre.item()
                else:
                    pre = pre.to_dense().tolist()
                all_predictions[idx] = pre
        return all_predictions

    @abc.abstractmethod
    def transform_predictions(self, batch: Batch) -> Batch:
        """
        Post-process the predictions of the torch model (e.g., apply
        threshold in case of multi-label classifier).
        """


class MulticlassTorchPredictor(TorchPredictor):
    """
    Applies a pre-trained torch model to unseen data for
    multiclass classification.
    """

    def transform_predictions(self, batch: Batch) -> Batch:
        batch = batch.argmax(-1)
        return batch


class MultilabelTorchPredictor(TorchPredictor):
    """
    Applies a pre-trained torch model to unseen data for
    multilabel classification, applying a threshold on the
    output of the neural network.
    """

    #: Threshold to transform probabilities into class predictions.
    #: Defaults to 0.5.
    threshold: float = 0.5

    def __init__(
        self,
        model: Union[nn.Module, ModelLoader],
        test_dataloader_class: str = 'torch.utils.data.DataLoader',
        test_dataloader_kwargs: Optional[Dict] = None,
        threshold: float = 0.5,
        name: str = None
    ) -> None:
        super().__init__(
            model, test_dataloader_class, test_dataloader_kwargs, name
        )
        self.threshold = threshold

    def transform_predictions(self, batch: Batch) -> Batch:
        return (batch > self.threshold).float()


class RegressionTorchPredictor(TorchPredictor):
    """
    Applies a pre-trained torch model to unseen data for
    regression, leaving untouched the output of the neural
    network.
    """

    def transform_predictions(self, batch: Batch) -> Batch:
        return batch



from torchvision import datasets, transforms


def imagenet_dataset(data_root: str):
    """Create a torch dataset object for Imagenet."""
    transform = transforms.Compose([
        transforms.Resize(256),
        transforms.RandomHorizontalFlip(),
        transforms.RandomVerticalFlip(),
        transforms.RandomRotation(degrees=45),
        transforms.ColorJitter(
            brightness=0.5, contrast=0.5, saturation=0.5, hue=0.5),
        transforms.CenterCrop(224),
        transforms.ToTensor(),
        transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ])
    imagenet = datasets.ImageFolder(
        root=data_root,
        transform=transform
    )
    return imagenet



from typing import Dict
import os
import shutil
import pickle
import random

import torch
from torch import Tensor
import matplotlib.pyplot as plt
import numpy as np

from itwinai.components import Saver, monitor_exec


class ParticleImagesSaver(Saver):
    """Saves generated particle trajectories to disk."""

    def __init__(
        self,
        save_dir: str = '3dgan-generated',
        aggregate_predictions: bool = False
    ) -> None:
        self.save_parameters(**self.locals2params(locals()))
        super().__init__()
        self.save_dir = save_dir
        self.aggregate_predictions = aggregate_predictions

    @monitor_exec
    def execute(self, generated_images: Dict[str, Tensor]) -> None:
        """Saves generated images to disk.

        Args:
            generated_images (Dict[str, Tensor]): maps unique item ID to
                the generated image.
        """
        if self.aggregate_predictions:
            os.makedirs(self.save_dir, exist_ok=True)
            sparse_generated_images = dict()
            for name, res in generated_images.items():
                sparse_generated_images[name] = res.to_sparse()
            del generated_images
            with open(self._random_file(), 'wb') as fp:
                pickle.dump(sparse_generated_images, fp)
        else:
            if os.path.exists(self.save_dir):
                shutil.rmtree(self.save_dir)
            os.makedirs(self.save_dir)
            # Save as torch tensor and jpg image
            for img_id, img in generated_images.items():
                img_path = os.path.join(self.save_dir, img_id)
                torch.save(img, img_path + '.pth')
                self._save_image(img, img_id, img_path + '.jpg')

    def _random_file(self, extension: str = 'pkl') -> str:
        fname = "%032x.%s" % (random.getrandbits(128), extension)
        fpath = os.path.join(self.save_dir, fname)
        while os.path.exists(fpath):
            fname = "%032x.%s" % (random.getrandbits(128), extension)
            fpath = os.path.join(self.save_dir, fname)
        return fpath

    def _save_image(
        self,
        img: Tensor,
        img_idx: str,
        img_path: str,
        center: bool = True
    ) -> None:
        """Converts a 3D tensor to a 3D scatter plot and saves it
        to disk as jpg image.
        """
        x_offset = img.shape[0] // 2 if center else 0
        y_offset = img.shape[1] // 2 if center else 0
        z_offset = img.shape[2] // 2 if center else 0

        # Convert tensor dimension IDs to coordinates
        x_coords = []
        y_coords = []
        z_coords = []
        values = []

        for x in range(img.shape[0]):
            for y in range(img.shape[1]):
                for z in range(img.shape[2]):
                    if img[x, y, z] > 0.0:
                        x_coords.append(x - x_offset)
                        y_coords.append(y - y_offset)
                        z_coords.append(z - z_offset)
                        values.append(img[x, y, z])

        # import plotly.graph_objects as go
        # normalize_intensity_by = 1
        # trace = go.Scatter3d(
        #     x=x_coords,
        #     y=y_coords,
        #     z=z_coords,
        #     mode='markers',
        #     marker_symbol='square',
        #     marker_color=[
        #         f"rgba(0,0,255,{i*100//normalize_intensity_by/10})"
        #         for i in values],
        # )
        # fig = go.Figure()
        # fig.add_trace(trace)
        # fig.write_image(img_path)

        values = np.array(values)
        # 0-1 scaling
        values = (values - values.min()) / (values.max() - values.min())
        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        ax.scatter(x_coords, y_coords, z_coords, alpha=values)
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        # Extract energy and angle from idx
        en, ang = img_idx.split('&')
        en = en[7:]
        ang = ang[6:]
        ax.set_title(f"Energy: {en} - Angle: {ang}")
        fig.savefig(img_path)



"""
Tests for MNIST use case.

Intended to be integration tests, to make sure that updates in the code base
do not break use cases' workflows.
"""

import pytest
import subprocess
import os

CYCLONES_PATH = "use-cases/cyclones"


@pytest.mark.skip("deprecated")
def test_structure_cyclones(check_folder_structure):
    """Test cyclones folder structure."""
    check_folder_structure(CYCLONES_PATH)


@pytest.mark.functional
@pytest.mark.memory_heavy
def test_cyclones_train_tf(tf_env, tmp_test_dir, install_requirements):
    """
    Test Cyclones tensorflow trainer by running it end-to-end.

    If CMCCC_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    # TODO: create a small sample dataset for tests only
    install_requirements(CYCLONES_PATH, tf_env)
    if os.environ.get('CMCCC_DATASET'):
        dataset_path = os.environ.get('CMCCC_DATASET')
    else:
        dataset_path = './data/tmp_data'
    pipe = os.path.join(os.path.abspath(CYCLONES_PATH), 'pipeline.yaml')
    train = os.path.join(os.path.abspath(CYCLONES_PATH), 'train.py')
    cmd = (f"{tf_env}/bin/python {train} "
           f"-p {pipe} --data_path {dataset_path}")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)



"""
Show how to use TensorFlow MultiWorkerMirroredStrategy on itwinai.

with SLURM:
>>> sbatch tfmirrored_slurm.sh

"""
from typing import Any
import argparse
import tensorflow as tf
from tensorflow import keras
from itwinai.tensorflow.distributed import get_strategy


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--strategy", "-s", type=str,
        choices=['mirrored'],
        default='mirrored'
    )
    parser.add_argument(
        "--batch_size", "-bs", type=int,
        default=64
    )
    parser.add_argument(
        "--shuffle_dataloader",
        action=argparse.BooleanOptionalAction
    )

    args = parser.parse_args()
    return args


def tf_rnd_dataset(args):
    """Dummy TF dataset."""
    (x_train, y_train), (x_test, y_test) = \
        tf.keras.datasets.mnist.load_data(
            path='p/scratch/intertwin/datasets/.keras/datasets/mnist.npz')

    train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))
    train_dataset = train_dataset.batch(args.batch_size)

    test_dataset = tf.data.Dataset.from_tensor_slices((x_test, y_test))
    test_dataset = test_dataset.batch(args.batch_size)

    return train_dataset, test_dataset


def trainer_entrypoint_fn(
        foo: Any, args: argparse.Namespace, strategy
) -> int:
    """Dummy training function, similar to custom code developed
    by some use case.
    """
    # dataset to be trained
    train_dataset, test_dataset = tf_rnd_dataset(args)

    # distribute datasets among mirrored replicas
    dist_train = strategy.experimental_distribute_dataset(
        train_dataset
    )
    dist_test = strategy.experimental_distribute_dataset(
        test_dataset
    )

    # define and compile model within strategy.scope()
    with strategy.scope():
        # Local model
        model = tf.keras.models.Sequential([
            tf.keras.layers.Flatten(input_shape=(28, 28)),
            tf.keras.layers.Dense(128, activation='relu'),
            tf.keras.layers.Dense(10)
        ])

        model.compile(loss=keras.losses.SparseCategoricalCrossentropy
                      (from_logits=True),
                      optimizer=keras.optimizers.RMSprop(),
                      metrics=['accuracy']
                      )

    model.fit(dist_train,
              epochs=5,
              steps_per_epoch=2000)

    test_scores = model.evaluate(dist_test, verbose=0, steps=500)

    print('Test loss:', test_scores[0])
    print('Test accuracy:', test_scores[1])

    return 123


if __name__ == "__main__":

    args = parse_args()

    # Instantiate Strategy
    if args.strategy == 'mirrored':
        if (len(tf.config.list_physical_devices('GPU')) == 0):
            raise RuntimeError('Resources unavailable')
        strategy, num_replicas = get_strategy()
    else:
        raise NotImplementedError(
            f"Strategy {args.strategy} is not recognized/implemented.")

    # Launch distributed training
    trainer_entrypoint_fn("foobar", args, strategy)




import torch
import numpy as np
from gwpy.timeseries import TimeSeries


def init_weights(net, init_type='normal', scaling=0.02):
    """
    Initialize the weights of the neural network according to the specified
    initialization type.

    Parameters:
        - net (nn.Module): The neural network model.
        - init_type (str): Type of initialization. Options: 'normal', 'xavier'
            (default is 'normal').
        - scaling (float): Scaling factor for weight initialization
            (default is 0.02).
    """
    def init_func(m):  # define the initialization function
        classname = m.__class__.__name__
        if hasattr(m, 'weight') and (classname.find('Conv')) != -1:
            torch.nn.init.normal_(m.weight.data, 0.0, scaling)
        # BatchNorm Layer's weight is not a matrix; only normal
        # distribution applies.
        elif classname.find('BatchNorm2d') != -1:
            torch.nn.init.normal_(m.weight.data, 1.0, scaling)
            torch.nn.init.constant_(m.bias.data, 0.0)

    print('initialize network with %s' % init_type)
    net.apply(init_func)  # apply the initialization function


def calculate_iou_2d(generated, target, threshold):
    """
    Calculate Intersection over Union (IoU) in the 2D plane at the specified
    intensity threshold.

    Parameters:
    - generated: List of time series representing the generated spectrograms
    - target: List of time series representing the target spectrograms
    - threshold: Intensity threshold for determining the binary masks

    Returns:
    - IoU: Intersection over Union
    """
    # Extract spectrogram values from time series
    # print(generated[0][0])
    # print(generated[0][0].shape)
    # print(type(generated[0][0]))

    spectrograms_gen = [TimeSeries(
        t[0], dt=1/4096.0).q_transform(frange=(10, 1000)).value
        for t in generated]
    spectrograms_real = [TimeSeries(
        t[0], dt=1/4096.0).q_transform(frange=(10, 1000)).value
        for t in target]

    # Create binary masks based on the intensity threshold
    mask1 = [spectrogram >= threshold for spectrogram in spectrograms_gen]
    mask2 = [spectrogram >= threshold for spectrogram in spectrograms_real]

    # Calculate the intersection and union of the binary masks
    intersection = [np.logical_and(m1, m2) for m1, m2 in zip(mask1, mask2)]
    union = [np.logical_or(m1, m2) for m1, m2 in zip(mask1, mask2)]

    # Calculate Intersection over Union (IoU)
    iou_list = np.array([np.sum(inter) / np.sum(uni)
                        for inter, uni in zip(intersection, union)])

    iou = iou_list.mean()
    return iou



from sklearn.preprocessing import MinMaxScaler
import tensorflow as tf
import joblib


def fit_transform(
    volume, shape, channel, feature_range=(0, 1),
        type='minmax', save=False, filename=None
):
    """
    Creates the scaler on the input volume and scales the data

    Parameters
    ----------
    volume : np.array
        Input 4-dimensional data volume.
    shape : (int, int)
        Height and width of the input volume.
    channel : int
        Number of channels of the input volume.
    feature_range : (int, int) | (0,1)
        Desired range of the scaled features. Default is (0,1).
    type : {'minmax', 'std', ...}
        Type of scaling. Default MinMax scaler
    save : bool | False
        Whether or not to save the computed scaler. Default to false
    filename : pathlike
        The filename to which the scaler must be saved to disk. Checked only
        if 'save' is set to True.

    Returns
    -------
    volume : np.array
        Scaled 4-dimensional input volume.
    scaler : scikit-learn scaler
        Computed scaler.
    """

    can_save = False
    if save:
        if filename is None:
            raise ValueError(
                'Must specify the filename when saving the scaler')
        else:
            can_save = True

    if type == 'minmax':
        scaler = MinMaxScaler(feature_range=feature_range)

    volume = scaler.fit_transform(
        volume.reshape(-1, channel)).reshape(-1, *shape, channel)

    if can_save:
        joblib.dump(scaler, filename)

    return volume, scaler


def inv_transform(scaled_image, scaler, shape, channel):
    """
    Scale the data from scaler's feature_range to data range

    Parameters
    ----------
    scaled_image : np.array
        4-dimensional input data volume.
    scaler : scikit-learn scaler
        Desired data scaler.
    shape : tuple
        (int, int) height and width of the input volume.
    channel : int
        Number of channels of the input volume.

    Returns
    -------
    image : np.array
        Scaled input volume.
    """
    return scaler.inverse_transform(
        scaled_image.reshape(-1, channel)
    ).reshape(*shape)


def transform(image, scaler, shape, channel):
    """
    Scale the data from data range to scaler's feature range

    Parameters
    ----------
    image : np.array
        4-dimensional input data volume.
    scaler : scikit-learn scaler
        Desired data scaler.
    shape : tuple
        (int, int) height and width of the input volume.
    channel : int
        Number of channels of the input volume.

    Returns
    -------
    scaled_image : np.array
        Scaled input volume.
    """
    return scaler.transform(
        image.reshape(-1, channel)
    ).reshape(-1, *shape, channel)


def get_scalers(scaler_X_file=None, scaler_y_file=None):
    """
    Reads the X and y scalers from file
    Input:
        @ (string) scaler_X_file : X data scaler of joblib.
            None if data must not be scaled
        @ (string) scaler_y_file : y data scaler of joblib.
            None if data must not be scaled
    Output:
        @ ([scaler,scaler]) scalers : loaded data scalers
    """
    X_scaler = None
    y_scaler = None
    if scaler_X_file:
        X_scaler = joblib.load(scaler_X_file)
    if scaler_y_file:
        y_scaler = joblib.load(scaler_y_file)
    return [X_scaler, y_scaler]


def save_tf_minmax(Xt, outfile):
    """
    Saves a MinMax Scaler as a Tensorflow Record.

    """
    def tensor_feature(value):
        """Returns a bytes_list from a string / byte."""
        return tf.train.Feature(bytes_list=tf.train.BytesList(
            value=[tf.io.serialize_tensor(tf.convert_to_tensor(value)).numpy()]
        )
        )

    def scaler_encoding_fn(min, max):
        """Builds a serialized version of the dataset. X and y
        must be np.array.
        """
        features = tf.train.Features(feature={
            "min": tensor_feature(min),
            "max": tensor_feature(max),
        })
        return tf.train.Example(features=features).SerializeToString()

    def write_record_to_file(min, max, record_file):
        with tf.io.TFRecordWriter(record_file) as writer:
            record = scaler_encoding_fn(min=min, max=max)
            writer.write(record)
        return

    # passed Xt is a numpy array of shape N x H x W x C
    batch_size = 8192
    n_batches = Xt.shape[0] // batch_size
    if Xt.shape[0] % batch_size:
        n_batches += 1

    # compute min
    cur_min = tf.math.reduce_min(input_tensor=Xt[0, ], axis=(0, 1)).numpy()
    for i in range(n_batches):
        X_batch = Xt[(i * batch_size):((i+1) * batch_size)]
        i_min = tf.math.reduce_min(
            input_tensor=X_batch, axis=(0, 1, 2)).numpy()
        for c in range(i_min.shape[-1]):
            if i_min[c] <= cur_min[c]:
                cur_min[c] = i_min[c]
    X_min = cur_min

    # compute max
    cur_max = tf.math.reduce_max(input_tensor=Xt[0, ], axis=(0, 1)).numpy()
    for i in range(n_batches):
        X_batch = Xt[(i * batch_size):((i+1) * batch_size)]
        i_max = tf.math.reduce_max(
            input_tensor=X_batch, axis=(0, 1, 2)).numpy()
        for c in range(i_max.shape[-1]):
            if i_max[c] >= cur_max[c]:
                cur_max[c] = i_max[c]
    X_max = cur_max

    # write min and max to file
    if outfile:
        write_record_to_file(min=X_min, max=X_max, record_file=outfile)

    # return scaler dictionary
    return {'min': tf.convert_to_tensor(X_min),
            'max': tf.convert_to_tensor(X_max)}


def save_tf_minmax_by_min_and_max(min, max, outfile):
    """
    Saves a MinMax Scaler as a Tensorflow Record.
    """
    def tensor_feature(value):
        """Returns a bytes_list from a string / byte."""
        return tf.train.Feature(bytes_list=tf.train.BytesList(value=[
            tf.io.serialize_tensor(tf.convert_to_tensor(value)).numpy()]))

    def scaler_encoding_fn(min, max):
        """Builds a serialized version of the dataset. X and y must be
        np.array.
        """
        features = tf.train.Features(feature={
            "min": tensor_feature(min),
            "max": tensor_feature(max),
        })
        return tf.train.Example(features=features).SerializeToString()

    def write_record_to_file(min, max, record_file):
        with tf.io.TFRecordWriter(record_file) as writer:
            record = scaler_encoding_fn(min=min, max=max)
            writer.write(record)
        return

    # write min and max to file
    if outfile:
        write_record_to_file(min=min, max=max, record_file=outfile)

    # return scaler dictionary
    return {'min': tf.convert_to_tensor(min), 'max': tf.convert_to_tensor(max)}


def load_tf_minmax(scalerfile, vars):
    """
    Loads a MinMax Scaler from disk as a Tensorflow Record.
    """
    AUTOTUNE = tf.data.AUTOTUNE

    def scaler_decoding_fn(serialized_data):
        """Decoding function for a dataset written to disk as
        tensor_encoding_fn().
        """
        features = {
            'min': tf.io.FixedLenFeature([], tf.string),
            'max': tf.io.FixedLenFeature([], tf.string)
        }
        # Parse the serialized data so we get a dict with our data.
        parsed_data = tf.io.parse_single_example(
            serialized_data, features=features)
        # Get X and y raw data
        raw_min = parsed_data['min']
        raw_max = parsed_data['max']
        # Decode the raw bytes so it becomes a tensor with type.
        min = tf.ensure_shape(tf.io.parse_tensor(
            raw_min, tf.float32), (len(vars)))
        max = tf.ensure_shape(tf.io.parse_tensor(
            raw_max, tf.float32), (len(vars)))
        return min, max

    # load scaler set
    scaler_set = (
        tf.data.TFRecordDataset(scalerfile, num_parallel_reads=AUTOTUNE)
        .map(scaler_decoding_fn, num_parallel_calls=AUTOTUNE)
    )
    # get min and max from the dataset
    for data in scaler_set:
        min, max = data

    # return scaler dictionary
    return {'min': min, 'max': max}


def minmax_transform(data, scaler):
    """
    Applies the transform of TFMinMaxScaler to the provided dataset.
    """
    if scaler:
        num = tf.subtract(data, scaler['min'])
        den = tf.subtract(scaler['max'], scaler['min'])
        res = tf.math.divide(num, den)
    else:
        res = data
    return res


def minmax_inverse_transform(scaled_data, scaler):
    """
    Applies the inverse transform of TFMinMaxScaler to the provided
    scaled dataset.
    """
    sub = tf.subtract(scaler['max'], scaler['min'])
    mul = tf.multiply(scaled_data, sub)
    return mul + scaler['min']


def minmax_inverse_target_transform(y_scaled, label_no_cyclone, patch_size):
    """
    Applies the inverse transform on y data when scaled in (0,1)
    """
    sub = tf.subtract(
        tf.cast(patch_size-1, dtype=tf.float32), label_no_cyclone)
    mul = tf.multiply(y_scaled, sub)
    return mul + label_no_cyclone



import torch
import torch.nn as nn


device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

# Mean-Squared Error as the average difference between the pixels
# in the original image vs. the reconstructed one
criterion = nn.MSELoss()
# pixel-wise MSE loss
pixel_wise_criterion = nn.MSELoss(reduction='none')

# KL divergence handles dispersion of information in latent space
# a balance is to be found with the prevailing reconstruction error
beta = 0.1

# number of evaluations for each dataset
n_avg = 20



#!/usr/bin/env python
# coding: utf-8

# ## Preprocess Data for seasonal VAEs

# The aim of this notebook is to translate numpy files (.npy) and time series (.csv)
# of daily maximum temperature to four numpy 3D-arrays: one for each season.
# These output arrays can easily be read for training and evaluating the
# Convolutional Variational AutoEncoder model.

# #### 0. Libraries

import numpy as np
import pandas as pd

from itwinai.components import DataGetter, monitor_exec

class SplitPreprocessedData(DataGetter):
    def __init__(
            self,
            scenario: str
    ):
        super().__init__()
        self.scenario = scenario

    ##### 2. Split Yearly Data into Four Seasonal Datasets
    # split daily data into seasons
    def season_split(
        self,
        images: np.ndarray,
        time: pd.DataFrame,
        dataset_type: str,
        n_memb: int,
        scenario: str = "",
    ) -> tuple[list[np.ndarray], list[pd.DataFrame]]:
        """
        Splits and returns the data sets (climate variable and time) per season.

        Parameters:
        images: temperature data sets
        time: time data sets
        dataset_type: 'train', 'test', 'proj'
        n_memb: number of ensemble members
        scenario: '' for train or test data,
                  or '126', '245', '370', '585' for projections
        """

        n_years = int(len(images) / 365)

        # 1st April = index 90
        # 1st July = index 181
        # 1st October = index 273
        winter_index = [365 * i + j for i in range(n_years) for j in range(90)]
        spring_index = [365 * i + j for i in range(n_years) for j in range(90, 181)]
        summer_index = [365 * i + j for i in range(n_years) for j in range(181, 273)]
        autumn_index = [365 * i + j for i in range(n_years) for j in range(273, 365)]

        winter_images = images[winter_index]
        spring_images = images[spring_index]
        summer_images = images[summer_index]
        autumn_images = images[autumn_index]

        winter_time = time.loc[winter_index].iloc[:, 1]
        spring_time = time.loc[spring_index].iloc[:, 1]
        summer_time = time.loc[summer_index].iloc[:, 1]
        autumn_time = time.loc[autumn_index].iloc[:, 1]

        # save results as an input for CVAE training
        np.save(
            f"input/preprocessed_1d_{dataset_type}{scenario}_winter_data_{n_memb}memb.npy",
            winter_images,
        )
        np.save(
            f"input/preprocessed_1d_{dataset_type}{scenario}_spring_data_{n_memb}memb.npy",
            spring_images,
        )
        np.save(
            f"input/preprocessed_1d_{dataset_type}{scenario}_summer_data_{n_memb}memb.npy",
            summer_images,
        )
        np.save(
            f"input/preprocessed_1d_{dataset_type}{scenario}_autumn_data_{n_memb}memb.npy",
            autumn_images,
        )
        pd.DataFrame(winter_time).to_csv(
            f"input/dates_{dataset_type}_winter_data_{n_memb}memb.csv"
        )
        pd.DataFrame(spring_time).to_csv(
            f"input/dates_{dataset_type}_spring_data_{n_memb}memb.csv"
        )
        pd.DataFrame(summer_time).to_csv(
            f"input/dates_{dataset_type}_summer_data_{n_memb}memb.csv"
        )
        pd.DataFrame(autumn_time).to_csv(
            f"input/dates_{dataset_type}_autumn_data_{n_memb}memb.csv"
        )

        season_images = [winter_images, spring_images, summer_images, autumn_images]
        season_time = winter_time, spring_time, summer_time, autumn_time

        return season_images, season_time

    @monitor_exec
    def execute(self):
        ##### 1. Load Data to xarray

        # choose the needed number of members
        n_memb = 1

        # define relevant scenarios
        #scenarios = ["126", "245", "370", "585"]
        scenarios = [self.scenario]
        # Load preprocessed "daily temperature images" and time series

        train_images = np.load("input/preprocessed_2d_train_data_allssp.npy")
        test_images = np.load("input/preprocessed_2d_test_data_allssp.npy")
        train_time = pd.read_csv("input/dates_train_data.csv")
        test_time = pd.read_csv("input/dates_test_data.csv")


        ##### 3. Apply to Train and Test Datasets
        train_season_images, train_season_time = self.season_split(
            train_images, train_time, "train", n_memb
        )

        test_season_images, test_season_time = self.season_split(
            test_images, test_time, "test", n_memb
        )


        ##### 4. Apply to Projection Datasets

        for scenario in scenarios:
            proj_images = np.load(f"input/preprocessed_2d_proj{scenario}_data_allssp.npy")
            proj_time = pd.read_csv("input/dates_proj_data.csv")

            proj_season_images, proj_season_time = self.season_split(
                proj_images, proj_time, "proj", n_memb, scenario
            )



#!/usr/bin/env python
# coding: utf-8

# ## Preprocess Data for VAE

# The aim of this notebook is to translate NetCDF files (.nc)
# of a daily climate variable (e.g. maximum temperature)
# to a numpy 3D-array. This output array can be read for training
# and evaluating the Convolutional Variational AutoEncoder model.


# #### 0. Libraries

import numpy as np
import xarray as xr
import cftime
import pandas as pd

from itwinai.components import DataGetter, monitor_exec

class PreprocessData(DataGetter):
    def __init__(
            self,
            scenario: str,
            dataset_root: str
    ):
        super().__init__()
        self.scenario = scenario
        self.dataset_root = dataset_root

    def xr_to_ndarray(self, xr_dset: xr.Dataset, sq_coords: dict) -> (np.ndarray, np.array, str):
        """
        Converts an xarray dataset it to a cropped square ndarray,
        after ajusting the longitudes from [0,360] to [-180,180].

        Parameters:
        xr_dset: data set of a climate variable
        sq_coords: spatial coordinates of the crop
        """
        # adjust the longitudes to keep a continuum over Europe
        xr_dset.coords["lon"] = (xr_dset.coords["lon"] + 180) % 360 - 180
        # re-organize data
        xr_dset = xr_dset.sortby(xr_dset.lon)
        # crop to the right square
        xr_dset = xr_dset.sel(
            lon=slice(sq_coords["min_lon"], sq_coords["max_lon"]),
            lat=slice(sq_coords["min_lat"], sq_coords["max_lat"]),
        )
        time_list = np.array(xr_dset["time"])
        n_t = len(time_list)
        n_lat = len(xr_dset.coords["lat"])
        n_lon = len(xr_dset.coords["lon"])
        nd_dset = np.ndarray((n_t, n_lat, n_lon, 1), dtype="float32")
        climate_variable = xr_dset.attrs["variable_id"]
        nd_dset[:, :, :, 0] = xr_dset[climate_variable][:, :, :]
        nd_dset = np.flip(nd_dset, axis=1)

        return nd_dset, time_list

    def sftlf_to_ndarray(
        self, xr_dset: xr.Dataset, sq_coords: dict
    ) -> (np.ndarray, np.array, str):
        """
        Converts and normalizes the land-sea mask data set
        to a cropped square ndarray,
        after ajusting the longitudes from [0,360] to [-180,180].

        Parameters:
        xr_dset: land-sea mask data set
        sq_coords: spatial coordinates of the crop
        """
        xr_dset.coords["lon"] = (xr_dset.coords["lon"] + 180) % 360 - 180
        xr_dset = xr_dset.sortby(xr_dset.lon)
        xr_dset = xr_dset.sel(
            lon=slice(sq_coords["min_lon"], sq_coords["max_lon"]),
            lat=slice(sq_coords["min_lat"], sq_coords["max_lat"]),
        )
        lat_list = xr_dset.coords["lat"]
        lon_list = xr_dset.coords["lon"]
        n_lat = len(lat_list)
        n_lon = len(lon_list)
        land_prop = np.ndarray((n_lat, n_lon, 1), dtype="float32")
        climate_variable = xr_dset.attrs["variable_id"]
        land_prop[:, :, 0] = xr_dset[climate_variable][:, :]
        # flip upside down to have North up
        land_prop = np.flipud(land_prop)
        # normalize data (originally in [0,100])
        land_prop = land_prop / 100

        return land_prop, lat_list, lon_list

    def get_extrema(self, histo_dataset: np.ndarray, proj_dataset: np.ndarray) -> np.array:
        """
        Computes global extrema over past and future data.

        Parameters:
        histo_dataset: historical data
        proj_dataset: projection data
        """
        global_min = min(np.min(histo_dataset), np.min(proj_dataset))
        global_max = max(np.max(histo_dataset), np.max(proj_dataset))
        return np.array([global_min, global_max])

    def normalize(self, nd_dset: np.ndarray, extrema: np.array) -> np.ndarray:
        """
        Normalizes a data set with given extrema.

        Parameters:
        nd_dset: data set of a climate variable
        extrema: extrema of the climate variable ([min, max])
        """
        norm_dset = (nd_dset - extrema[0]) / (extrema[1] - extrema[0])
        return norm_dset

    ##### 4. Split Historical Data into Train and Test Datasets
    # Train the network on most of the historical data,
    # but keep some to test the model performance on new data points.
    def split_train_test(
        self, nd_dset: np.ndarray, time_list: np.array, train_proportion: float = 0.8
    ) -> (np.ndarray, np.ndarray, np.array, np.array):
        """
        Splits a data set into train and test data (and time).

        Parameters:
        nd_dset: data set to be split
        time_list: time list to be split
        train_proportion: proportion of train data
        """
        len_train = int(len(nd_dset) * train_proportion)
        train_data = nd_dset[:len_train]
        test_data = nd_dset[len_train:]
        train_time = time_list[:len_train]
        test_time = time_list[len_train:]
        return train_data, test_data, train_time, test_time

    def split_date(
        self, nd_dset: np.ndarray, time_list: np.array, year: int = 1950
    ) -> (np.ndarray, np.ndarray, np.array, np.array):
        """
        Splits a data set into train and test data (and time),
        if the previous train_proportion splits data in the middle of a year.

        Parameters:
        nd_dset: data set to be split
        time_list: time list to be split
        year: year where the data is to be split
        """
        split_index = np.where(
            time_list == cftime.DatetimeNoLeap(year, 1, 1, 0, 0, 0, 0, has_year_zero=True)
        )[0][0]
        train_data = nd_dset[:split_index]
        test_data = nd_dset[split_index:]
        train_time = time_list[:split_index]
        test_time = time_list[split_index:]
        return train_data, test_data, train_time, test_time

    ##### 5. Combine into a 2D-Array
    def ndarray_to_2d(self, temp_dset: np.ndarray, land_prop: np.ndarray) -> np.ndarray:
        """
        Combines climate variable and land-sea mask data sets.

        Parameters:
        temp_dset: climate variable data
        land_prop: land-sea mask data
        """
        n_t = np.shape(temp_dset)[0]
        n_lat = np.shape(temp_dset)[1]
        n_lon = np.shape(temp_dset)[2]

        # combine all variables on a same period to a new 2D-array
        total_dset = np.zeros((n_t, n_lat, n_lon, 2), dtype="float32")
        total_dset[:, :, :, 0] = temp_dset.reshape(n_t, n_lat, n_lon)
        total_dset[:, :, :, 1] = np.transpose(
            np.repeat(land_prop, n_t, axis=2), axes=[2, 0, 1]
        )

        return total_dset

    @monitor_exec
    def execute(self):
        #### 1. Load Data to xarrays
        data_dir = self.dataset_root

        # Historical Datasets
        # regrouped by climate variable

        temp_50 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_historical_r1i1p1f1_gn_19500101-19741231.nc"
        )
        temp_75 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_historical_r1i1p1f1_gn_19750101-19991231.nc"
        )
        temp_00 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_historical_r1i1p1f1_gn_20000101-20141231.nc"
        )
        temp_histo = xr.concat([temp_50, temp_75, temp_00], "time")

        # Projection Datasets
        # regrouped by climate variable
        # IPCC scenarios: SSP1-2.6, SSP2-4.5, SSP3-7.0, SSP5-8.5
        # choose among "126", "245", "370", "585"
        scenario = self.scenario

        temp_15 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20150101-20391231.nc"
        )
        temp_40 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20400101-20641231.nc"
        )
        temp_65 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20650101-20891231.nc"
        )
        temp_90 = xr.open_dataset(
            f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20900101-21001231.nc"
        )
        temp_proj = xr.concat([temp_15, temp_40, temp_65, temp_90], "time")

        # Load land-sea mask data
        sftlf = xr.open_dataset(
            f"{data_dir}sftlf_fx_CESM2_historical_r9i1p1f1_gn.nc",
            chunks={"time": 10},
        )

        ##### 2. Restrict to a Geospatial Square
        sq32_west_europe = {"min_lon": -10, "max_lon": 29, "min_lat": 36, "max_lat": 66}

        land_prop, lat_list, lon_list = self.sftlf_to_ndarray(sftlf, sq32_west_europe)

        ##### 6. Step-by-Step Preprocessing

        # Crop original data to Western Europe
        temp_histo_nd, time_list = self.xr_to_ndarray(temp_histo, sq32_west_europe)
        temp_proj_nd, time_proj = self.xr_to_ndarray(temp_proj, sq32_west_europe)


        # Compute the variable extrema over history and projections
        # temp_extrema = get_extrema(temp_histo_nd, temp_proj_nd)

        # Here are the results for CMCC-ESM2 (all scenarios)
        # to save time
        temp_extrema = np.array([234.8754, 327.64])
        # ssp585 array([234.8754, 327.64  ], dtype=float32)
        # ssp370 array([234.8754 , 325.43323], dtype=float32)
        # ssp245 array([234.8754, 324.8263], dtype=float32)
        # ssp126 array([234.8754, 323.6651], dtype=float32)

        # Normalize data with the above extrema
        temp_histo_norm = self.normalize(temp_histo_nd, temp_extrema)
        temp_proj_norm = self.normalize(temp_proj_nd, temp_extrema)

        # Split data into train and test data sets
        train_temp, test_temp, train_time, test_time = self.split_train_test(
            temp_histo_norm, time_list
        )

        # Add up split data and land-sea mask
        total_train = self.ndarray_to_2d(train_temp, land_prop)
        total_test = self.ndarray_to_2d(test_temp, land_prop)
        total_proj = self.ndarray_to_2d(temp_proj_norm, land_prop)


        ##### 7. Save Results

        # Save train and test data sets
        np.save("input/preprocessed_2d_train_data_allssp.npy", total_train)
        np.save("input/preprocessed_2d_test_data_allssp.npy", total_test)
        pd.DataFrame(train_time).to_csv("input/dates_train_data.csv")
        pd.DataFrame(test_time).to_csv("input/dates_test_data.csv")

        # Save projection data for one scenario
        np.save(f"input/preprocessed_2d_proj{scenario}_data_allssp.npy", total_proj)
        pd.DataFrame(time_proj).to_csv("input/dates_proj_data.csv")


        ##### 8. Preprocessing for All Scenarios

        # This part is to be run as a complement to 6. and 7.

        # Here you can remove the scenario you already run in 6. and 7.
        #scenarios = ["126", "245", "370", "585"]
        #TODO: Discuss with Anne/Christian
        #scenarios = ["245", "585"]
        scenarios = [self.scenario]

        for scenario in scenarios:

            # Load projection data
            temp_15 = xr.open_dataset(
                f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20150101-20391231.nc"
            )
            temp_40 = xr.open_dataset(
                f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20400101-20641231.nc"
            )
            temp_65 = xr.open_dataset(
                f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20650101-20891231.nc"
            )
            temp_90 = xr.open_dataset(
                f"{data_dir}tasmax_day_CMCC-ESM2_ssp{scenario}_r1i1p1f1_gn_20900101-21001231.nc"
            )
            temp_proj = xr.concat([temp_15, temp_40, temp_65, temp_90], "time")

            # Process projection data
            temp_proj_nd, time_proj = self.xr_to_ndarray(temp_proj, sq32_west_europe)
            # extrema for all scenarios in CMCC-ESM2
            temp_extrema = np.array([234.8754, 327.64])
            temp_proj_norm = self.normalize(temp_proj_nd, temp_extrema)
            total_proj = self.ndarray_to_2d(temp_proj_norm, land_prop)

            # Save results
            np.save(f"input/preprocessed_2d_proj{scenario}_data_allssp.npy", total_proj)



from typing import Literal, Optional
import os
import torch.nn as nn
import torch
import time
import numpy as np

from itwinai.torch.trainer import TorchTrainer
from itwinai.torch.distributed import (
    DeepSpeedStrategy,
)
from itwinai.torch.config import TrainingConfiguration
from itwinai.loggers import Logger

from src.model import Decoder, Decoder_2d_deep, UNet, GeneratorResNet
from src.utils import init_weights


from tqdm import tqdm


class NoiseGeneratorTrainer(TorchTrainer):

    def __init__(
        self,
        batch_size: int,
        learning_rate: float = 1e-3,
        num_epochs: int = 2,
        generator: Literal["simple", "deep", "resnet", "unet"] = "unet",
        loss: Literal["L1", "L2"] = "L1",
        strategy: Literal["ddp", "deepspeed", "horovod"] = 'ddp',
        checkpoint_path: str = "checkpoints/epoch_{}.pth",
        save_best: bool = True,
        logger: Optional[Logger] = None,
        random_seed: Optional[int] = None,
        name: str | None = None
    ) -> None:
        super().__init__(
            epochs=num_epochs,
            config={},
            strategy=strategy,
            logger=logger,
            random_seed=random_seed,
            name=name
        )
        self.save_parameters(**self.locals2params(locals()))
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self._generator = generator
        self._loss = loss
        self.checkpoint_path = checkpoint_path
        os.makedirs(os.path.dirname(checkpoint_path), exist_ok=True)
        # Global training configuration
        self.config = TrainingConfiguration(
            batch_size=batch_size,
            save_best=save_best,
            shuffle_train=True
        )

    def create_model_loss_optimizer(self) -> None:
        # Select generator
        generator = self._generator.lower()
        if generator == "simple":
            self.model = Decoder(3, norm=False)
            init_weights(self.model, 'normal', scaling=.02)
        elif generator == "deep":
            self.model = Decoder_2d_deep(3)
            init_weights(self.model, 'normal', scaling=.02)
        elif generator == "resnet":
            self.model = GeneratorResNet(3, 12, 1)
            init_weights(self.model, 'normal', scaling=.01)
        elif generator == "unet":
            self.model = UNet(
                input_channels=3, output_channels=1, norm=False)
            init_weights(self.model, 'normal', scaling=.02)
        else:
            raise ValueError("Unrecognized generator type! Got", generator)

        # Select loss
        loss = self._loss.upper()
        if loss == "L1":
            self.loss = nn.L1Loss()
        elif loss == "L2":
            self.loss = nn.MSELoss()
        else:
            raise ValueError("Unrecognized loss type! Got", loss)

        # Optimizer
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate)

        # IMPORTANT: model, optimizer, and scheduler need to be distributed

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(
                    train_micro_batch_size_per_gpu=self.config.batch_size
                )
            )
        else:
            distribute_kwargs = {}

        # Distributed model, optimizer, and scheduler
        self.model, self.optimizer, _ = self.strategy.distributed(
            self.model, self.optimizer, **distribute_kwargs
        )

    def train(self):
        # uncomment all lines relative to accuracy if you want to measure
        # IOU between generated and real spectrograms.
        # Note that it significantly slows down the whole process
        # it also might not work as the function has not been fully
        # implemented yet

        loss_plot = []
        val_loss_plot = []
        acc_plot = []
        val_acc_plot = []
        best_val_loss = float('inf')
        for epoch in tqdm(range(self.num_epochs)):
            # itwinai - IMPORTANT: set current epoch ID
            self.set_epoch(epoch)

            st = time.time()
            epoch_loss = []
            # epoch_acc = []
            for i, batch in enumerate(self.train_dataloader):
                # batch= transform(batch)
                target = batch[:, 0].unsqueeze(1).to(self.device)
                # print(f'TARGET ON DEVICE: {target.get_device()}')
                target = target.float()
                input = batch[:, 1:].to(self.device)
                # print(f'INPUT ON DEVICE: {input.get_device()}')

                self.optimizer.zero_grad()
                generated = self.model(input.float())
                # generated=normalize_(generated,1)
                loss = self.loss(generated, target)
                loss.backward()
                self.optimizer.step()
                epoch_loss.append(loss.detach().cpu().numpy())
                # itwinai - log loss as metric
                self.log(loss.detach().cpu().numpy(),
                         'epoch_loss_batch',
                         kind='metric',
                         step=epoch*len(self.train_dataloader) + i,
                         batch_idx=i)
                # acc=accuracy(generated.detach().cpu().numpy(),target.detach().cpu().numpy(),20)
                # epoch_acc.append(acc)
            val_loss = []
            # val_acc = []
            for i, batch in enumerate(self.validation_dataloader):
                # batch= transform(batch)
                target = batch[:, 0].unsqueeze(1).to(self.device)
                target = target.float()
                input = batch[:, 1:].to(self.device)
                with torch.no_grad():
                    generated = self.model(input.float())
                    # generated=normalize_(generated,1)
                    loss = self.loss(generated, target)
                val_loss.append(loss.detach().cpu().numpy())
                # itwinai -log loss as metric
                self.log(loss.detach().cpu().numpy(),
                         'val_loss_batch',
                         kind='metric',
                         step=epoch*len(self.validation_dataloader) + i,
                         batch_idx=i)
                # acc=accuracy(generated.detach().cpu().numpy(),target.detach().cpu().numpy(),20)
                # val_acc.append(acc)
            loss_plot.append(np.mean(epoch_loss))
            val_loss_plot.append(np.mean(val_loss))
            # acc_plot.append(np.mean(epoch_acc))
            # val_acc_plot.append(np.mean(val_acc))

            # itwinai - Log metrics/losses
            self.log(np.mean(epoch_loss), 'epoch_loss',
                     kind='metric', step=epoch)
            self.log(np.mean(val_loss), 'val_loss',
                     kind='metric', step=epoch)
            # self.log(np.mean(epoch_acc), 'epoch_acc',
            #          kind='metric', step=epoch)
            # self.log(np.mean(val_acc), 'val_acc',
            #          kind='metric', step=epoch)

            # print('epoch: {} loss: {} val loss: {} accuracy: {} val
            # accuracy: {}'.format(epoch,loss_plot[-1],val_loss_plot[-1],
            # acc_plot[-1],val_acc_plot[-1]))
            et = time.time()
            # itwinai - print() in a multi-worker context (distributed)
            if self.strategy.is_main_worker:
                print('epoch: {} loss: {} val loss: {} time:{}s'.format(
                    epoch, loss_plot[-1], val_loss_plot[-1], et-st))

            # Save checkpoint every 100 epochs
            if epoch % 1 == 0:
                # uncomment the following if you want to save checkpoint every
                # 100 epochs regardless of the performance of the model
                # checkpoint = {
                #     'epoch': epoch,
                #     'model_state_dict': generator.state_dict(),
                #     'optim_state_dict': optimizer.state_dict(),
                #     'loss': loss_plot[-1],
                #     'val_loss': val_loss_plot[-1],
                # }
                # if self.strategy.is_main_worker:
                #     # Save only in the main worker
                #     checkpoint_filename = checkpoint_path.format(epoch)
                #     torch.save(checkpoint, checkpoint_filename)

                # Average loss among all workers
                # itwinai - gather local loss from all the workers
                worker_val_losses = self.strategy.gather_obj(val_loss_plot[-1])
                if self.strategy.is_main_worker:
                    # Save only in the main worker

                    # avg_loss has a meaning only in the main worker
                    avg_loss = np.mean(worker_val_losses)

                    # instead of val_loss and best_val loss we should
                    # use accuracy!!!
                    if self.config.save_best and avg_loss < best_val_loss:
                        # create checkpoint
                        checkpoint = {
                            'epoch': epoch,
                            'model_state_dict': self.model.state_dict(),
                            'optim_state_dict': self.optimizer.state_dict(),
                            'loss': loss_plot[-1],
                            'val_loss': val_loss_plot[-1],
                        }

                        # save checkpoint only if it is better than
                        # the previous ones
                        checkpoint_filename = self.checkpoint_path.format(
                            epoch)
                        torch.save(checkpoint, checkpoint_filename)
                        # itwinai - log checkpoint as artifact
                        self.log(checkpoint_filename,
                                 os.path.basename(checkpoint_filename),
                                 kind='artifact')

                        # update best model
                        best_val_loss = val_loss_plot[-1]
                        best_checkpoint_filename = (
                            self.checkpoint_path.format('best')
                        )
                        torch.save(checkpoint, best_checkpoint_filename)
                        # itwinai - log checkpoint as artifact
                        self.log(best_checkpoint_filename,
                                 os.path.basename(best_checkpoint_filename),
                                 kind='artifact')
        # return (loss_plot, val_loss_plot,
        # acc_plot, val_acc_plot ,acc_plot, val_acc_plot)
        return loss_plot, val_loss_plot, acc_plot, val_acc_plot



import abc
from typing import Any, List, Optional, Tuple, Union, Iterable, Literal
from pathlib import Path
import json
import os

import deepspeed
import torch
import torch.distributed as dist
import horovod.torch as hvd
import torch.nn as nn
import torch.optim as optim
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data import Dataset, Sampler, DistributedSampler, DataLoader
from torch.utils.data.dataloader import T_co, _worker_init_fn_t, _collate_fn_t

from ..distributed import DistributedStrategy
from .type import UninitializedStrategyError, DistributedStrategyError


def distributed_resources_available() -> bool:
    """Check if the current execution environment
    has (enough) GPUs available to allow for distributed ML.

    Returns:
        bool: env can support distributed ML.
    """
    if torch.cuda.is_available() and torch.cuda.device_count() > 1:
        return True
    return False


class TorchDistributedStrategy(DistributedStrategy):
    """Abstract class to define the distributed backend methods for
    PyTorch models.
    """

    #: Allows to discriminate distributed strategies from non-distributed.
    #: Defaults to True.
    is_distributed: bool = True
    #: Set to True when the current strategy is initialized.
    #: Defaults to False.
    is_initialized: bool = False

    @property
    def is_main_worker(self) -> bool:
        """Checks if local worker has global rank equal to zero.

        Returns:
            bool: True if main worker.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return self.global_rank() == 0

    @abc.abstractmethod
    def init(self) -> None:
        """Initializes the chosen distributed backend"""
    # @abc.abstractmethod
    # def distributed_engine(
    #     self, model: nn.Module, optimizer: Optimizer,
    #     lr_scheduler: Optional[LRScheduler] = None
    # ) -> ModelEngine:
    #     """Build a distributed model engine."""

    @abc.abstractmethod
    def distributed(
        self, model: nn.Module, optimizer: Optimizer,
        lr_scheduler: Optional[LRScheduler] = None
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""

    @abc.abstractmethod
    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """

    @abc.abstractmethod
    def local_world_size(self) -> int:
        """Returns the number of local workers available on a node
        (local world size).
        Usually it is equal to the number of available GPUs.

        Returns:
            int: local world size.
        """

    @abc.abstractmethod
    def global_rank(self) -> int:
        """Returns the global rank of the current process.
        Rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """

    @abc.abstractmethod
    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """

    def device(self) -> str:
        """Device used by local worker.

        Returns:
            str: torch device in the form 'cuda:N'.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return f"cuda:{self.local_rank()}"

    def set_device(self):
        """Set local device."""
        torch.cuda.device(self.local_rank())
        # Needed by torch.distributed.gather_object
        torch.cuda.set_device(self.local_rank())

    def create_dataloader(
        self, dataset: Dataset[T_co], batch_size: Optional[int] = 1,
        shuffle: Optional[bool] = None,
        sampler: Union[Sampler, Iterable, None] = None,
        batch_sampler: Union[Sampler[List], Iterable[List], None] = None,
        num_workers: int = 0, collate_fn: Optional[_collate_fn_t] = None,
        pin_memory: bool = False, drop_last: bool = False,
        timeout: float = 0,
        worker_init_fn: Optional[_worker_init_fn_t] = None,
        multiprocessing_context=None, generator=None,
        *, prefetch_factor: Optional[int] = None,
        persistent_workers: bool = False,
        pin_memory_device: str = ""
    ):
        """Create a distributed DataLoader by using ``DistributedSampler`` as
        random sampler.

        Args:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: ``1``).
            shuffle (bool, optional): set to ``True`` to have the data
                reshuffled at every epoch (default: ``False``).
            sampler (Sampler or Iterable, optional): defines the strategy to
                draw
                samples from the dataset. Can be any ``Iterable`` with
                ``__len__``
                implemented. If specified, :attr:`shuffle` must not be
                specified.
            batch_sampler (Sampler or Iterable, optional): like
                :attr:`sampler`, but
                returns a batch of indices at a time. Mutually exclusive with
                :attr:`batch_size`, :attr:`shuffle`, :attr:`sampler`,
                and :attr:`drop_last`.
            num_workers (int, optional): how many subprocesses to use for data
                loading. ``0`` means that the data will be loaded in the main
                process. (default: ``0``)
            collate_fn (Callable, optional): merges a list of samples to form a
                mini-batch of Tensor(s).  Used when using batched loading from
                a map-style dataset.
            pin_memory (bool, optional): If ``True``, the data loader will
                copy Tensors
                into device/CUDA pinned memory before returning them.  If your
                data elements
                are a custom type, or your :attr:`collate_fn` returns a batch
                that is a custom type,
                see the example below.
            drop_last (bool, optional): set to ``True`` to drop the last
                incomplete batch,
                if the dataset size is not divisible by the batch size.
                If ``False`` and
                the size of dataset is not divisible by the batch size, then
                the last batch
                will be smaller. (default: ``False``)
            timeout (numeric, optional): if positive, the timeout value for
                collecting a batch
                from workers. Should always be non-negative. (default: ``0``)
            worker_init_fn (Callable, optional): If not ``None``,
                this will be called on each
                worker subprocess with the worker id (an int in
                ``[0, num_workers - 1]``) as
                input, after seeding and before data loading.
                (default: ``None``)
            multiprocessing_context (str or
                multiprocessing.context.BaseContext, optional): If
                ``None``, the default `multiprocessing context`_ of
                your operating system will
                be used. (default: ``None``)
            generator (torch.Generator, optional): If not ``None``,
                this RNG will be used
                by RandomSampler to generate random indexes and
                multiprocessing to generate
                ``base_seed`` for workers. (default: ``None``)
            prefetch_factor (int, optional, keyword-only arg): Number of
                batches loaded
                in advance by each worker. ``2`` means there will be a total of
                2 * num_workers batches prefetched across all workers.
                (default value depends
                on the set value for num_workers. If value of num_workers=0
                default is ``None``.
                Otherwise, if value of ``num_workers > 0`` default is ``2``).
            persistent_workers (bool, optional): If ``True``, the data loader
                will not shut down
                the worker processes after a dataset has been consumed once.
                This allows to
                maintain the workers `Dataset` instances alive.
                (default: ``False``)
            pin_memory_device (str, optional): the device to
                :attr:`pin_memory` to if ``pin_memory`` is ``True``.


        .. warning:: If the ``spawn`` start method is used,
                    :attr:`worker_init_fn`
                    cannot be an unpicklable object, e.g., a lambda function.
                    See `Multiprocessing best practices`_ on more
                    details related to multiprocessing in PyTorch.


        .. warning:: ``len(dataloader)`` heuristic is based on the length of
                    the sampler used.
                    When :attr:`dataset` is an
                    :class:`~torch.utils.data.IterableDataset`,
                    it instead returns an estimate based on
                    ``len(dataset) / batch_size``, with proper
                    rounding depending on :attr:`drop_last`, regardless
                    of multi-process loading
                    configurations. This represents the best guess PyTorch
                    can make because PyTorch
                    trusts user :attr:`dataset` code in correctly handling
                    multi-process
                    loading to avoid duplicate data.

                    However, if sharding results in multiple workers having
                    incomplete last batches,
                    this estimate can still be inaccurate, because (1) an
                    otherwise complete batch can
                    be broken into multiple ones and (2) more than one batch
                    worth of samples can be
                    dropped when :attr:`drop_last` is set. Unfortunately,
                    PyTorch can not detect such cases in general.

                    See `Dataset Types`_ for more details on these two
                    types of datasets and how
                    :class:`~torch.utils.data.IterableDataset` interacts with
                    `Multi-process data loading`_.


        .. warning:: See `Reproducibility`_, and
                    `My data loader workers return identical random numbers`_,
                    and
                    `Randomness in multi-process data loading`_ notes for
                    random seed related questions.


        .. _multiprocessing context:
            https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        .. _Multiprocessing best practices:
            https://pytorch.org/docs/stable/notes/multiprocessing.html#multiprocessing-best-practices
        .. _Reproducibility:
            https://pytorch.org/docs/stable/notes/randomness.html#reproducibility
        .. _My data loader workers return identical random numbers:
            https://pytorch.org/docs/stable/notes/faq.html#dataloader-workers-random-seed
        .. _Randomness in multi-process data loading:
            https://pytorch.org/docs/stable/data.html#data-loading-randomness
        .. _Multi-process data loading:
            https://pytorch.org/docs/stable/data.html#multi-process-data-loading
        .. _Dataset Types:
            https://pytorch.org/docs/stable/data.html#dataset-types
    """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")

        if self.is_distributed:
            if sampler is not None:
                raise RuntimeError(
                    "User-provided sampler is not supported."
                )
            sampler = DistributedSampler(
                dataset, num_replicas=self.global_world_size(),
                rank=self.global_rank(),
                shuffle=shuffle
            )
        # shuffle and batch_sampler must be unset
        return DataLoader(
            dataset=dataset, batch_size=batch_size, sampler=sampler,
            num_workers=num_workers, collate_fn=collate_fn,
            pin_memory=pin_memory, drop_last=drop_last, timeout=timeout,
            worker_init_fn=worker_init_fn,
            multiprocessing_context=multiprocessing_context,
            generator=generator, prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
            pin_memory_device=pin_memory_device
        )

    @abc.abstractmethod
    def clean_up(self) -> None:
        """Cleans up resources allocated by distributed strategy."""

    @abc.abstractmethod
    def allgather_obj(self, obj: Any) -> List[Any]:
        """All-gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.

        Returns:
            List[Any]: list of objects gathered from all workers.
        """

    @abc.abstractmethod
    def gather_obj(self, obj: Any, dst_rank: int = 0) -> List[Any]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            List[Any]: list of objects gathered from all workers.
        """


class TorchDDPStrategy(TorchDistributedStrategy):
    """PyTorch ``DistributedDataParallel`` distributed strategy class.

    Args:
        backend (Literal['nccl', 'gloo', 'mpi']): Name of the
            distributed communication backend to employ.
    """

    #: Torch distributed communication backend.
    backend: Literal['nccl', 'gloo', 'mpi']

    def __init__(self, backend: Literal['nccl', 'gloo', 'mpi']) -> None:
        super().__init__()
        self.backend = backend

    def init(self) -> None:
        """Initializes the distributed process group and the distributed
        package.

        Raises:
            RuntimeError: when there are not (enough) GPUs available.
            DistributedStrategyError: when trying to initialize a strategy
                which is already initialized.
        """
        if not distributed_resources_available():
            raise RuntimeError(
                "Trying to run distributed on insufficient resources.")
        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")
        dist.init_process_group(backend=self.backend)
        self.is_initialized = True

        self.set_device()

    # def distributed_engine(
    #     self, model: nn.Module, optimizer: Optimizer,
    #     lr_scheduler: Optional[LRScheduler] = None,
    #     mixed_precision: bool = False
    # ) -> ModelEngine:
    #     """Build a distributed model engine."""
    #     if torch.cuda.is_available():
    #         # device = self.dist_lrank()
    #         model = model.to(self.dist_device())
    #         dist_model = torch.nn.parallel.DistributedDataParallel(
    #             model,
    #             device_ids=[self.dist_device()],
    #             output_device=self.dist_device()
    #         )
    #     else:
    #         dist_model = model

    #     model_engine = DDPModelEngine(
    #         dist_model, optimizer, lr_scheduler,
    #         mixed_precision=mixed_precision
    #     )

    #     return model_engine

    def distributed(
        self, model: nn.Module, optimizer: Optimizer,
        lr_scheduler: Optional[LRScheduler] = None,
        **kwargs
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        if torch.cuda.is_available():
            # device = self.dist_lrank()
            model = model.to(self.device())
            dist_model = torch.nn.parallel.DistributedDataParallel(
                model,
                device_ids=[self.device()],
                output_device=self.device()
            )
        else:
            dist_model = model

        return dist_model, optimizer, lr_scheduler

    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return dist.get_world_size()

    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return torch.cuda.device_count()

    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return dist.get_rank()

    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return dist.get_rank() % torch.cuda.device_count()

    def clean_up(self) -> None:
        """Destroys the current process group."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        if torch.cuda.is_available():
            dist.barrier()
            dist.destroy_process_group()

    def allgather_obj(self, obj: Any) -> List[Any]:
        """All-gathers any object from the whole group
        in a list (to all workers).

        Args:
            obj (Any): Object to gather from all workers.

        Returns:
            List[Any]: List of gathered objects.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        res = [None] * self.global_world_size()
        dist.all_gather_object(res, obj)
        return res

    def gather_obj(self, obj: Any, dst_rank: int = 0) -> Optional[List[Any]]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[Any]]: list of objects gathered from all workers
            or ``None`` on non-destination ranks.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        if self.global_rank() == dst_rank:
            res = [None] * self.global_world_size()
            dist.gather_object(obj, res, dst=dst_rank)
            return res

        dist.gather_object(obj, dst=dst_rank)


class DeepSpeedStrategy(TorchDistributedStrategy):
    """DeepSpeed distributed strategy class.

    Args:
        backend (Literal['nccl', 'gloo', 'mpi']): Name of the
            distributed communication backend to employ.
        config (Union[dict, Path, str]): DeepSpeed config. Either a
            dictionary or a path to a JSON file.
    """

    #: Torch distributed communication backend.
    backend: Literal['nccl', 'gloo', 'mpi']

    def __init__(
        self,
        backend: Literal['nccl', 'gloo', 'mpi']
    ) -> None:
        super().__init__()
        self.backend = backend

    def _load_config(self, ds_config) -> None:
        if isinstance(ds_config, (str, Path)):
            with open(ds_config) as fp:
                self.config = json.load(fp)
        elif isinstance(ds_config, dict):
            self.config = ds_config
        else:
            raise ValueError("ds_config is neither a dictionary not a path.")

    def init(self) -> None:
        """Initializes the distributed process group and the distributed
        package.

        Raises:
            RuntimeError: when there are not (enough) GPUs available.
            DistributedStrategyError: when trying to initialize a strategy
                already initialized.
        """
        if not distributed_resources_available():
            raise RuntimeError(
                "Trying to run distributed on insufficient resources.")

        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")

        # https://github.com/Lightning-AI/pytorch-lightning/issues/13567
        ompi_lrank = os.environ.get('OMPI_COMM_WORLD_LOCAL_RANK')
        os.environ['OMPI_COMM_WORLD_LOCAL_RANK'] = os.environ.get(
            'LOCAL_RANK', ompi_lrank)

        # https://deepspeed.readthedocs.io/en/latest/initialize.html#training-initialization
        deepspeed.init_distributed(dist_backend=self.backend)
        self.is_initialized = True

        self.set_device()

    def distributed(
        self, model: nn.Module, optimizer: Optional[Optimizer] = None,
        lr_scheduler: Optional[LRScheduler] = None,
        model_parameters: Optional[Any] = None,
        **init_kwargs
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")

        if init_kwargs.get("config"):
            self._load_config(init_kwargs.get("config"))
        # https://deepspeed.readthedocs.io/en/latest/initialize.html#training-initialization
        # To prioritize optim in the config, you need to pass optim=None
        distrib_model, optimizer, _, lr_scheduler = deepspeed.initialize(
            model=model,
            model_parameters=model_parameters,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            dist_init_required=True,
            **init_kwargs
        )
        return distrib_model, optimizer, lr_scheduler

    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return dist.get_world_size()

    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return torch.cuda.device_count()

    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return dist.get_rank()

    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return dist.get_rank() % torch.cuda.device_count()

    def clean_up(self) -> None:
        """Destroys the current process group."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        # deepspeed.sys.exit() # disabled as it kills the execution

    def allgather_obj(self, obj: Any) -> List[Any]:
        """All-gathers any object from the whole group
        in a list (to all workers).

        Args:
            obj (Any): Object to gather from all workers.

        Returns:
            List[Any]: List of gathered objects.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        res = [None] * self.global_world_size()
        dist.all_gather_object(res, obj)
        return res

    def gather_obj(self, obj: Any, dst_rank: int = 0) -> Optional[List[Any]]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[Any]]: list of objects gathered from all workers
            or ``None`` on non-destination ranks.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        if self.global_rank() == dst_rank:
            res = [None] * self.global_world_size()
            dist.gather_object(obj, res, dst=dst_rank)
            return res

        dist.gather_object(obj, dst=dst_rank)


class HorovodStrategy(TorchDistributedStrategy):
    """Horovod distributed strategy class."""

    def init(self) -> None:
        """Initializes the Horovod distributed backend.

        Raises:
            RuntimeError: when there are not (enough) GPUs available.
            DistributedStrategyError: when trying to initialize a strategy
                already initialized.
        """
        if not distributed_resources_available():
            raise RuntimeError(
                "Trying to run distributed on insufficient resources.")
        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")
        hvd.init()
        self.is_initialized = True

        self.set_device()

    def distributed(
        self, model: nn.Module, optimizer: Optional[Optimizer] = None,
        lr_scheduler: Optional[LRScheduler] = None,
        **optim_kwargs
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")

        model.to(self.device())

        # Scale learning rate
        # https://github.com/horovod/horovod/issues/1653#issuecomment-574764452
        lr_scaler = 1
        if optim_kwargs.get('op') == hvd.Adasum:
            lr_scaler = hvd.local_size()
        elif optim_kwargs.get('op') == hvd.Average:
            lr_scaler = hvd.size()
        for g in optimizer.param_groups:
            g['lr'] *= lr_scaler

        self._broadcast_params(model, optimizer)

        distOptimizer = hvd.DistributedOptimizer(
            optimizer,
            named_parameters=model.named_parameters(),
            **optim_kwargs
        )
        return model, distOptimizer, lr_scheduler

    def _broadcast_params(
            self, model: nn.Module, optimizer: optim.Optimizer
    ) -> None:
        """Broadcasts variables from root rank to all other processes.

        Args:
            model (nn.Module): ML model that is to be broadcasted
                across processes.
            optimizer (optim.Optimizer): Optimizer that is to be broadcasted
                across processes.
        """
        hvd.broadcast_parameters(model.state_dict(), root_rank=0)
        hvd.broadcast_optimizer_state(optimizer, root_rank=-0)

    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return hvd.size()

    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return hvd.local_size()

    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return hvd.rank()

    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return hvd.local_rank()

    def clean_up(self) -> None:
        """Shuts Horovod down."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        hvd.shutdown()

    def allgather_obj(self, obj: Any) -> list[Any]:
        """All-gathers scalar objects across all workers to a
        list with size(#worker), uses horovod communicator

        Args:
            obj (Any): object in a worker.

        Returns:
            list: gathered list with size(#worker).
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return hvd.allgather_object(obj)

    def gather_obj(self, obj: Any, dst_rank: int = 0) -> list[Any]:
        """The same as ``allgather_obj``, as gather is not supported
        by Horovod.

        Args:
            obj (Any): object in a worker.
            dst_rank (int): ignored.

        Returns:
            list: gathered list with size(#worker).
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        return self.allgather_obj(obj)


class NonDistributedStrategy(TorchDistributedStrategy):
    """Dummy class for non-distributed environments."""

    #: This strategy is not distributed.
    #: Defaults to False.
    is_distributed: bool = True
    is_distributed: bool = False

    def init(self) -> None:
        """If CUDA is available set CUDA device, and do nothing more.

        Raises:
            DistributedStrategyError: when trying to initialize a strategy
                already initialized.
        """
        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")
        if torch.cuda.is_available():
            self.set_device()
        self.is_initialized = True

    def device(self) -> str:
        """Device used by local worker.

        Returns:
            str: cpu device if CUDA is not available.
        """
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        if torch.cuda.is_available():
            return super().device()
        return "cpu"

    def distributed(
        self, model: nn.Module, optimizer: Optional[Optimizer] = None,
        lr_scheduler: Optional[LRScheduler] = None,
        **kwargs
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Do nothing and return model, optimizer and scheduler."""
        if not self.is_initialized:
            raise UninitializedStrategyError(
                "Strategy has not been initialized. Use the init method.")
        if torch.cuda.is_available():
            model = model.cuda()
        return model, optimizer, lr_scheduler

    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        return 1

    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.
        """
        return 1

    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        return 0

    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        return 0

    def clean_up(self) -> None:
        """Do nothing."""

    def allgather_obj(self, obj: Any) -> list[Any]:
        """Wraps ``obj`` into a List object.

        Args:
            obj (Any): object in a worker.

        Returns:
            list[Any]: input object wrapped in a list.
        """
        return [obj]

    def gather_obj(self, obj: Any, dst_rank: int = 0) -> list[Any]:
        """Wraps ``obj`` into a List object.

        Args:
            obj (Any): object in a worker.
            dst_rank (int): ignored.

        Returns:
            list[Any]: input object wrapped in a list.
        """
        return [obj]



"""
In this section we define different NN architectures models, and initialise
one of them as the generator to use in training and inference.

This section is split in three parts:
- 1) **Weight Initialization**, where we define the function to initialise the
    weights of the NN models according to certain parameters and distributions
    given as input.
- 2) **NN Models**, where we define different NN models exploiting different
    architectures.
- 3) **Generator**, where we initialise one of the above models as the
    generator to use in training and inference
"""

import torch.nn as nn
import torch


# SHALLOW DECODER


class Decoder(nn.Module):
    """
    Decoder network.

    Args:
        - in_channels (int): Number of input channels.
        - kernel_size (int): Size of the convolutional kernel (default is 7).
        - a (float): Scaling factor (default is 80.0).
        - norm (bool): Whether to apply normalization (default is True).
    """

    def __init__(self, in_channels, kernel_size=7, norm=True):
        super(Decoder, self).__init__()

        self.conv1 = nn.Conv2d(
            in_channels, 32, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        self.relu1 = nn.LeakyReLU(0.2, inplace=True)

        self.conv2 = nn.Conv2d(
            32, 64, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        self.relu2 = nn.LeakyReLU(0.2, inplace=True)

        self.conv3 = nn.Conv2d(
            64, 64, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        self.relu3 = nn.LeakyReLU(0.2, inplace=True)

        self.conv4 = nn.Conv2d(
            64, 1, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)

        if norm:
            self.activation = torch.nn.Sigmoid()
        else:
            self.activation = torch.nn.ReLU()

    def _forward_features(self, x):
        """
        Perform forward pass through the network layers.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        x = self.relu1(self.conv1(x))
        x = self.relu2(self.conv2(x))
        x = self.relu3(self.conv3(x))
        x = self.conv4(x)
        x = self.activation(x)
        return x

    def forward(self, x):
        """
        Forward pass through the network.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        return self._forward_features(x)


# DEEP DECODER


class Decoder_2d_deep(nn.Module):
    """
    Deep 2D decoder network.

    Args:
        - in_channels (int): Number of input channels.
        - kernel_size (int): Size of the convolutional kernel (default is 5).
    """

    def __init__(self, in_channels, kernel_size=5, norm=True):
        super(Decoder_2d_deep, self).__init__()

        self.conv1 = nn.Conv2d(
            in_channels, 64, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        self.relu1 = nn.LeakyReLU(0.3, inplace=True)

        self.conv2 = nn.Conv2d(
            64, 128, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        self.relu2 = nn.LeakyReLU(0.3, inplace=True)

        self.conv3 = nn.Conv2d(
            128, 256, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        self.relu3 = nn.LeakyReLU(0.3, inplace=True)

        self.conv4 = nn.Conv2d(
            256, 1, kernel_size=kernel_size, stride=1,
            padding=kernel_size // 2)
        if norm:
            self.activation = torch.nn.Sigmoid()
        else:
            self.activation = torch.nn.ReLU()

    def _forward_features(self, x):
        """
        Perform forward pass through the network layers.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        x = self.relu1(self.conv1(x))
        x = self.relu2(self.conv2(x))
        x = self.relu3(self.conv3(x))
        x = self.conv4(x)
        x = self.activation(x)
        return x

    def forward(self, x):
        """
        Forward pass through the network.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        return self._forward_features(x)


# RESNET


class ResidualBlock(nn.Module):
    """
    Residual block module.

    Args:
        - in_features (int): Number of input features/channels.
    """

    def __init__(self, in_features):
        super(ResidualBlock, self).__init__()

        # Define the block sequence
        self.block = nn.Sequential(
            # Pads the input tensor using the reflection of the input boundary
            nn.ReflectionPad2d(1),
            nn.Conv2d(in_features, in_features, 3),  # 2D convolutional layer
            nn.InstanceNorm2d(in_features),  # Instance normalization
            nn.ReLU(inplace=True),  # ReLU activation function
            nn.ReflectionPad2d(1),
            nn.Conv2d(in_features, in_features, 3),
            nn.InstanceNorm2d(in_features)
        )

    def forward(self, x):
        """
        Forward pass through the residual block.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        return x + self.block(x)


class GeneratorResNet(nn.Module):
    """
    Generator network using ResNet architecture.

    Args:
        - input_shape (int): Number of input features/channels.
        - num_residual_block (int): Number of residual blocks.
        - output_shape (int): Number of output features/channels.
    """

    def __init__(
            self, input_shape, num_residual_block, output_shape, norm=False
    ):
        super(GeneratorResNet, self).__init__()

        channels = input_shape
        target_channels = output_shape
        # Initial Convolution Block
        out_features = 64
        if norm:
            self.final_activation = torch.nn.Sigmoid()
        else:
            self.final_activation = torch.nn.ReLU()

        model = [
            nn.ReflectionPad2d(channels),
            nn.Conv2d(channels, out_features, 7),
            nn.InstanceNorm2d(out_features),
            nn.ReLU(inplace=True)
        ]
        in_features = out_features

        # Downsampling
        for _ in range(2):
            out_features *= 2
            model += [
                nn.Conv2d(in_features, out_features, 3, stride=2, padding=1),
                nn.InstanceNorm2d(out_features),
                nn.ReLU(inplace=True)
            ]
            in_features = out_features

        # Residual blocks
        for _ in range(num_residual_block):
            model += [ResidualBlock(out_features)]

        # Upsampling
        for _ in range(2):
            out_features //= 2
            model += [
                nn.Upsample(scale_factor=2),  # Upsampling layer
                nn.Conv2d(in_features, out_features, 3, stride=1, padding=1),
                nn.ReLU(inplace=True)
            ]
            in_features = out_features

        # Output Layer
        model += [nn.ReflectionPad2d(target_channels),
                  nn.Conv2d(out_features, target_channels, 3),
                  self.final_activation  # Sigmoid activation function
                  ]

        # Unpacking
        self.model = nn.Sequential(*model)

    def forward(self, x):
        """
        Forward pass through the generator network.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        return self.model(x)


# U-NET


class Conv2dBlock(nn.Module):
    """
    Convolutional block module.

    Args:
        - in_channels (int): Number of input channels.
        - out_channels (int): Number of output channels.
        - kernel_size (int): Size of the convolutional kernel (default is 3).
    """

    def __init__(self, in_channels, out_channels, kernel_size=3):
        super(Conv2dBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels, out_channels,
                               kernel_size, padding=1)
        self.conv2 = nn.Conv2d(
            out_channels, out_channels, kernel_size, padding=1)
        self.activation = nn.ReLU()

    def forward(self, x):
        """
        Forward pass through the convolutional block.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        x = self.conv1(x)
        x = self.activation(x)
        x = self.conv2(x)
        x = self.activation(x)
        return x


class EncoderBlock(nn.Module):
    """
    Encoder block module.

    Args:
        - in_channels (int): Number of input channels.
        - out_channels (int): Number of output channels.
        - pool_size (tuple): Size of the pooling kernel (default is (2, 2)).
        - dropout (float): Dropout rate (default is 0.3).
    """

    def __init__(self, in_channels, out_channels, pool_size=(2, 2),
                 dropout=0.3):
        super(EncoderBlock, self).__init__()
        self.conv_block = Conv2dBlock(in_channels, out_channels)
        self.maxpool = nn.MaxPool2d(pool_size)
        self.dropout = nn.Dropout2d(dropout)

    def forward(self, x):
        """
        Forward pass through the encoder block.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        features = self.conv_block(x)
        pooled = self.maxpool(features)
        pooled = self.dropout(pooled)
        return features, pooled


class UNetEncoder(nn.Module):
    """
    Encoder network module.

    Args:
        - input_channels (int): Number of input channels.
    """

    def __init__(self, input_channels):
        super(UNetEncoder, self).__init__()
        self.block1 = EncoderBlock(input_channels, 64)
        self.block2 = EncoderBlock(64, 128)
        self.block3 = EncoderBlock(128, 256)
        self.block4 = EncoderBlock(256, 512)

    def forward(self, x):
        """
        Forward pass through the encoder network.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        f1, p1 = self.block1(x)
        f2, p2 = self.block2(p1)
        f3, p3 = self.block3(p2)
        f4, p4 = self.block4(p3)
        return p4, (f1, f2, f3, f4)


class Bottleneck(nn.Module):
    """
    Bottleneck module.

    """

    def __init__(self):
        super(Bottleneck, self).__init__()
        self.conv_block = Conv2dBlock(512, 1024)

    def forward(self, x):
        """
        Forward pass through the bottleneck module.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        return self.conv_block(x)


class DecoderBlock(nn.Module):
    """
    Decoder block module.

    Args:
        - in_channels (int): Number of input channels.
        - out_channels (int): Number of output channels.
        - kernel_size (int): Size of the convolutional kernel (default is 3).
        - stride (int): Stride size for the convolutional operation
            (default is 2).
        - dropout (float): Dropout rate (default is 0.3).
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=2,
                 dropout=0.3):
        super(DecoderBlock, self).__init__()
        self.deconv = nn.ConvTranspose2d(
            in_channels, out_channels, kernel_size, stride=stride, padding=1,
            output_padding=1)
        self.dropout = nn.Dropout2d(dropout)
        self.conv_block = Conv2dBlock(out_channels * 2, out_channels)

    def forward(self, x, conv_output):
        """
        Forward pass through the decoder block.

        Args:
            - x (torch.Tensor): Input tensor.
            - conv_output (torch.Tensor): Output tensor from the corresponding
            encoder block.

        Returns:
            - torch.Tensor: Output tensor.
        """
        x = self.deconv(x, output_size=conv_output.size())
        x = torch.cat([x, conv_output], dim=1)
        x = self.dropout(x)
        x = self.conv_block(x)
        return x


class UNetDecoder(nn.Module):
    """
    Decoder network module.

    Args:
        - output_channels (int): Number of output channels.
    """

    def __init__(self, output_channels):
        super(UNetDecoder, self).__init__()
        self.block6 = DecoderBlock(1024, 512)
        self.block7 = DecoderBlock(512, 256)
        self.block8 = DecoderBlock(256, 128)
        self.block9 = DecoderBlock(128, 64)
        self.final_conv = nn.Conv2d(64, output_channels, kernel_size=1)

    def forward(self, x, convs):
        """
        Forward pass through the decoder network.

        Args:
            - x (torch.Tensor): Input tensor.
            - convs (tuple): Tuple containing the output tensors from the
                encoder blocks.

        Returns:
            - torch.Tensor: Output tensor.
        """
        f1, f2, f3, f4 = convs
        x = self.block6(x, f4)
        x = self.block7(x, f3)
        x = self.block8(x, f2)
        x = self.block9(x, f1)
        outputs = self.final_conv(x)

        return outputs


class UNet(nn.Module):
    """
    UNet network module.

    Args:
        - input_channels (int): Number of input channels.
        - output_channels (int): Number of output channels.
    """

    def __init__(self, input_channels=3, output_channels=1, norm=True):
        super(UNet, self).__init__()
        self.encoder = UNetEncoder(input_channels)
        self.bottleneck = Bottleneck()
        self.decoder = UNetDecoder(output_channels)
        if norm:
            self.final_activation = torch.nn.Sigmoid()
        else:
            self.final_activation = torch.nn.ReLU()

    def forward(self, x):
        """
        Forward pass through the UNet network.

        Args:
            - x (torch.Tensor): Input tensor.

        Returns:
            - torch.Tensor: Output tensor.
        """
        encoder_output, convs = self.encoder(x)
        bottleneck_output = self.bottleneck(encoder_output)
        outputs = self.decoder(bottleneck_output, convs)
        return self.final_activation(outputs)



"""
Training pipeline. To run this script, use the following commands.

On login node:

>>> python train.py -p pipeline.yaml -d

On compute nodes:

>>>  python train.py -p pipeline.yaml

"""

from typing import Dict
import argparse
import logging
from os.path import join
from os import makedirs
from datetime import datetime

# # the mock-0.3.1 dir contains testcase.py, testutils.py & mock.py
from itwinai.parser import ConfigParser, ArgumentParser

from src.macros import PATCH_SIZE, SHAPE


def dynamic_config(args) -> Dict:
    """Generates a configuration with values computed at runtime.

    Args:
        args (argparse.Namespace): arguments parsed from command line.

    Returns:
        Dict: configuration.
    """
    config = {}

    # Paths, Folders
    FORMATTED_DATETIME = str(datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
    MODEL_BACKUP_DIR = join(args.root_dir, "models/")
    EXPERIMENTS_DIR = join(args.root_dir, "experiments")
    RUN_DIR = join(EXPERIMENTS_DIR, args.run_name +
                   "_" + FORMATTED_DATETIME)
    SCALER_DIR = join(RUN_DIR, "scalers")
    TENSORBOARD_DIR = join(RUN_DIR, "tensorboard")
    CHECKPOINTS_DIR = join(RUN_DIR, "checkpoints")

    # Files
    LOG_FILE = join(RUN_DIR, "run.log")

    # Create folders
    makedirs(EXPERIMENTS_DIR, exist_ok=True)
    makedirs(RUN_DIR, exist_ok=True)
    makedirs(SCALER_DIR, exist_ok=True)
    makedirs(TENSORBOARD_DIR, exist_ok=True)
    makedirs(CHECKPOINTS_DIR, exist_ok=True)

    config = {
        "root_dir": args.root_dir,
        "experiment_dir": EXPERIMENTS_DIR,
        "run_dir": RUN_DIR,
        "scaler_dir": SCALER_DIR,
        "tensorboard_dir": TENSORBOARD_DIR,
        "checkpoints_dir": CHECKPOINTS_DIR,
        "backup_dir": MODEL_BACKUP_DIR,
        "log_file": LOG_FILE,
        "shape": SHAPE,
        "patch_size": PATCH_SIZE,
        # "epochs": args.epochs,
        # "batch_size": args.batch_size
    }

    # initialize logger
    logging.basicConfig(
        format="[%(asctime)s] %(levelname)s : %(message)s",
        level=logging.DEBUG,
        filename=LOG_FILE,
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    return config


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "-p", "--pipeline", type=str, required=True,
        help='Configuration file to the pipeline to execute.'
    )
    parser.add_argument("-r", "--root_dir", type=str, default='./data')
    parser.add_argument("--data_path", type=str,
                        default='./data/data_path')
    parser.add_argument("-n", "--run_name", default="noname", type=str)
    parser.add_argument(
        '-d', '--download-only',
        action=argparse.BooleanOptionalAction,
        default=False,
        help=('Whether to download only the dataset and exit execution '
              '(suggested on login nodes of HPC systems)')
    )

    args = parser.parse_args()
    global_config = dynamic_config(args)

    # Create parser for the pipeline
    pipe_parser = ConfigParser(
        config=args.pipeline,
        override_keys={
            "dataset_root": args.data_path,
            "global_config": global_config
        }
    )
    pipeline = pipe_parser.parse_pipeline(
        pipeline_nested_key='training_pipeline'
    )

    if args.download_only:
        print('Downloading datasets and exiting...')
        pipeline = pipeline[:1]

    pipeline.execute()



"""
This module provides the functionalities to execute workflows defined in
in form of pipelines.
"""
from __future__ import annotations
from typing import Iterable, Dict, Any, Tuple, Union, Optional

from .components import BaseComponent, monitor_exec
from .utils import SignatureInspector


class Pipeline(BaseComponent):
    """Executes a set of components arranged as a pipeline."""

    #: Pipeline steps. Can be a list of ``BaseComponent`` or a dictionary
    #: allowing the user to name each ``BaseComponent``.
    steps: Union[Dict[str, BaseComponent], Iterable[BaseComponent]]

    def __init__(
        self,
        steps: Union[Dict[str, BaseComponent], Iterable[BaseComponent]],
        name: Optional[str] = None
    ):
        super().__init__(name=name)
        self.save_parameters(steps=steps, name=name)
        self.steps = steps

    def __getitem__(self, subscript: Union[str, int, slice]) -> Pipeline:
        if isinstance(subscript, slice):
            # First, convert to list if is a dict
            if isinstance(self.steps, dict):
                steps = list(self.steps.items())
            else:
                steps = self.steps
            # Second, perform slicing
            s = steps[subscript.start:subscript.stop: subscript.step]
            # Third, reconstruct dict, if it is a dict
            if isinstance(self.steps, dict):
                s = dict(s)
            # Fourth, return sliced sub-pipeline, preserving its
            # initial structure
            sliced = self.__class__(
                steps=s,
                name=self.name
            )
            return sliced
        else:
            return self.steps[subscript]

    def __len__(self) -> int:
        return len(self.steps)

    @monitor_exec
    def execute(self, *args) -> Any:
        """"Execute components sequentially."""
        if isinstance(self.steps, dict):
            steps = list(self.steps.values())
        else:
            steps = self.steps

        for step in steps:
            step: BaseComponent
            args = self._pack_args(args)
            self.validate_args(args, step)
            args = step.execute(*args)

        return args

    @staticmethod
    def _pack_args(args) -> Tuple:
        """Wraps args in a tuple, if needed."""
        args = () if args is None else args
        if not isinstance(args, tuple):
            args = (args,)
        return args

    @staticmethod
    def validate_args(input_args: Tuple, component: BaseComponent):
        """Verify that the number of input args provided to some component
        match with the number of the non-default args in the component.

        Args:
            input_args (Tuple): input args to be fed to the component.
            component (BaseComponent): component to be executed.

        Raises:
            RuntimeError: in case of args mismatch.
        """
        inspector = SignatureInspector(component.execute)
        if inspector.min_params_num > len(input_args):
            raise TypeError(
                f"Component '{component.name}' received too few "
                f"input arguments: {input_args}. Expected at least "
                f"{inspector.min_params_num}, with names: "
                f"{inspector.required_params}."
            )
        if (inspector.max_params_num != inspector.INFTY
                and len(input_args) > inspector.max_params_num):
            raise TypeError(
                f"Component '{component.name}' received too many "
                f"input arguments: {input_args}. Expected at most "
                f"{inspector.max_params_num}."
            )



"""Create a simple inference dataset sample and a checkpoint."""

import torch
import os
import argparse

from model import Net
from dataloader import InferenceMNIST


def mnist_torch_inference_files(
    root: str = '.',
    samples_path: str = 'mnist-sample-data/',
    model_name: str = 'mnist-pre-trained.pth'
):
    """Create sample dataset and fake model to test mnist
    inference workflow. Assumes to be run from
    the use case folder.

    Args:
        root (str, optional): where to create the files.
        Defaults to '.'.
    """

    sample = os.path.join(root, samples_path)
    InferenceMNIST.generate_jpg_sample(sample, 10)

    # Fake checkpoint
    dummy_nn = Net()
    mdl_ckpt = os.path.join(root, model_name)
    torch.save(dummy_nn, mdl_ckpt)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", type=str, default='.')
    parser.add_argument("--samples-path", type=str,
                        default='mnist-sample-data')
    parser.add_argument("--model-name", type=str,
                        default='mnist-pre-trained.pth')
    args = parser.parse_args()
    mnist_torch_inference_files(**vars(args))



from typing import Optional, Tuple, Any
import os
import pandas as pd
import torch
from sklearn.model_selection import train_test_split
from itwinai.components import (
    DataGetter, DataProcessor, DataSplitter, monitor_exec
)

from src.dataset import (
    generate_dataset_aux_channels,
    generate_dataset_main_channel,
    generate_cut_image_dataset,
    normalize_
)


class TimeSeriesDatasetGenerator(DataGetter):
    # TODO: move configuration to the constructor.
    def __init__(
        self,
        data_root: str = "data",
        name: Optional[str] = None
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.data_root = data_root
        if not os.path.exists(data_root):
            os.makedirs(data_root, exist_ok=True)

    @monitor_exec
    def execute(self) -> pd.DataFrame:
        """Generate a time-series dataset, convert it to Q-plots,
        save it to disk, and return it.

        Returns:
            pd.DataFrame: dataset of Q-plot images.
        """
        df_aux_ts = generate_dataset_aux_channels(
            1000, 3, duration=16, sample_rate=500,
            num_waves_range=(20, 25), noise_amplitude=0.6
        )
        df_main_ts = generate_dataset_main_channel(
            df_aux_ts, weights=None, noise_amplitude=0.1
        )

        # save datasets
        save_name_main = 'TimeSeries_dataset_synthetic_main.pkl'
        save_name_aux = 'TimeSeries_dataset_synthetic_aux.pkl'
        df_main_ts.to_pickle(os.path.join(self.data_root, save_name_main))
        df_aux_ts.to_pickle(os.path.join(self.data_root, save_name_aux))

        # Transform to images and save to disk
        df_ts = pd.concat([df_main_ts, df_aux_ts], axis=1)
        df = generate_cut_image_dataset(
            df_ts, list(df_ts.columns),
            num_processes=20, square_size=64
        )
        save_name = 'Image_dataset_synthetic_64x64.pkl'
        df.to_pickle(os.path.join(self.data_root, save_name))
        return df


class TimeSeriesDatasetSplitter(DataSplitter):
    def __init__(
        self,
        train_proportion: int | float,
        validation_proportion: int | float = 0.0,
        test_proportion: int | float = 0.0,
        rnd_seed: int | None = None,
        images_dataset: str = "data/Image_dataset_synthetic_64x64.pkl",
        name: str | None = None
    ) -> None:
        super().__init__(
            train_proportion, validation_proportion,
            test_proportion, name
        )
        self.save_parameters(**self.locals2params(locals()))
        self.validation_proportion = 1-train_proportion
        self.rnd_seed = rnd_seed
        self.images_dataset = images_dataset

    def get_or_load(self, dataset: Optional[pd.DataFrame] = None):
        """If the dataset is not given, load it from disk."""
        if dataset is None:
            print("WARNING: loading time series dataset from disk.")
            return pd.read_pickle(self.images_dataset)
        return dataset

    @monitor_exec
    def execute(
        self,
        dataset: Optional[pd.DataFrame] = None
    ) -> Tuple:
        """Splits a dataset into train, validation and test splits.

        Args:
            dataset (pd.DataFrame): input dataset.

        Returns:
            Tuple: tuple of train, validation and test splits. Test is None.
        """
        dataset = self.get_or_load(dataset)

        # Convert data to torch
        df = dataset.applymap(lambda x: torch.tensor(x))

        # Divide Image dataset in main and aux channels. Note that df
        # generated in the section Generate Synthetic Dataset will always have
        # the main channel as its first column
        main_channel = list(df.columns)[0]
        aux_channels = list(df.columns)[1:]

        df_aux_all_2d = pd.DataFrame(df[aux_channels])
        df_main_all_2d = pd.DataFrame(df[main_channel])
        X_train_2d, X_test_2d, y_train_2d, y_test_2d = train_test_split(
            df_aux_all_2d, df_main_all_2d,
            test_size=self.validation_proportion, random_state=self.rnd_seed)
        return (X_train_2d, y_train_2d), (X_test_2d, y_test_2d), None


class TimeSeriesProcessor(DataProcessor):
    def __init__(self, name: str | None = None) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))

    @monitor_exec
    def execute(
        self,
        train_dataset: Tuple,
        validation_dataset: Tuple,
        test_dataset: Any = None
    ) -> Tuple[torch.Tensor, torch.Tensor, None]:
        """Pre-process datasets: rearrange and normalize before training.

        Args:
            train_dataset (Tuple): training dataset.
            validation_dataset (Tuple): validation dataset.
            test_dataset (Any, optional): unused placeholder. Defaults to None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, None]: train, validation, and
                test (placeholder) datasets. Ready to be used for training.
        """
        X_train_2d, y_train_2d = train_dataset
        X_test_2d, y_test_2d = validation_dataset

        # Name of the main channel (assuming it's in position 0)
        main_channel = list(y_train_2d.columns)[0]

        # TRAINING SET

        # # smaller dataset
        # signal_data_train_small_2d = torch.stack([
        #     torch.stack([y_train_2d[main_channel].iloc[i]])
        #     for i in range(100)
        # ])  # for i in range(y_train.shape[0])
        # aux_data_train_small_2d = torch.stack([
        #     torch.stack([X_train_2d.iloc[i][0], X_train_2d.iloc[i]
        #                 [1], X_train_2d.iloc[i][2]])
        #     for i in range(100)
        # ])  # for i in range(X_train.shape[0])

        # whole dataset
        signal_data_train_2d = torch.stack([
            torch.stack([y_train_2d[main_channel].iloc[i]])
            for i in range(y_train_2d.shape[0])
        ])
        aux_data_train_2d = torch.stack([
            torch.stack(
                [X_train_2d.iloc[i][0], X_train_2d.iloc[i][1],
                 X_train_2d.iloc[i][2]])
            for i in range(X_train_2d.shape[0])
        ])

        # concatenate torch.tensors
        train_data_2d = torch.cat(
            [signal_data_train_2d, aux_data_train_2d], dim=1)
        # train_data_small_2d = torch.cat(
        #     [signal_data_train_small_2d, aux_data_train_small_2d], dim=1)

        # VALIDATION SET

        # # smaller dataset
        # signal_data_test_small_2d = torch.stack([
        #     torch.stack(
        #         [y_test_2d[main_channel].iloc[i]])
        #     for i in range(100)
        # ])  # for i in range(y_test.shape[0])
        # aux_data_test_small_2d = torch.stack([
        #     torch.stack(
        #         [X_test_2d.iloc[i][0], X_test_2d.iloc[i][1],
        #          X_test_2d.iloc[i][2]])
        #     for i in range(100)
        # ])  # for i in range(X_test.shape[0])

        # whole dataset
        signal_data_test_2d = torch.stack([
            torch.stack(
                [y_test_2d[main_channel].iloc[i]])
            for i in range(y_test_2d.shape[0])
        ])
        aux_data_test_2d = torch.stack([
            torch.stack(
                [X_test_2d.iloc[i][0], X_test_2d.iloc[i][1],
                 X_test_2d.iloc[i][2]])
            for i in range(X_test_2d.shape[0])
        ])

        test_data_2d = torch.cat(
            [signal_data_test_2d, aux_data_test_2d], dim=1)
        # test_data_small_2d = torch.cat(
        #     [signal_data_test_small_2d, aux_data_test_small_2d], dim=1)

        # NORMALIZE
        train_data_2d = normalize_(train_data_2d)
        test_data_2d = normalize_(test_data_2d)

        return train_data_2d, test_data_2d, None
